@@ -1,5 +1,8 @@
 import sys
+import time
+from pyasn1.compat.octets import octs2ints
 from pysnmp import error
+from pysnmp import __version__
 
 flagNone     = 0x0000
 flagIO       = 0x0001
@@ -33,17 +36,26 @@ class Debug:
     def __init__(self, *flags):
         self._flags = flagNone
         self._printer = self.defaultPrinter
+        self('running pysnmp version %s' % __version__)
         for f in flags:
-            if f not in flagMap:
+            inverse = f and f[0] in ('!', '~')
+            if inverse:
+                f = f[1:]
+            try:
+                if inverse:
+                    self._flags &= ~flagMap[f]
+                else:
+                    self._flags |= flagMap[f]
+            except KeyError:
                 raise error.PySnmpError('bad debug flag %s' % f)
-            self._flags = self._flags | flagMap[f]
-            self('debug category %s enabled' % f)
+  
+            self('debug category \'%s\' %s' % (f, inverse and 'disabled' or 'enabled'))
         
     def __str__(self):
         return 'logger %s, flags %x' % (self._printer, self._flags)
     
     def __call__(self, msg):
-        self._printer('DBG: %s\n' % msg)
+        self._printer('DBG: [%s]: %s\n' % (self.timestamp(), msg))
 
     def __and__(self, flag):
         return self._flags & flag
@@ -51,6 +63,10 @@ class Debug:
     def __rand__(self, flag):
         return flag & self._flags
 
+    def timestamp(self):
+        return time.strftime('%H:%M:%S', time.localtime()) + \
+               '.%.3d' % int((time.time() % 1) * 1000)
+
 # This will yield false from bitwise and with a flag, and save
 # on unnecessary calls
 logger = 0
@@ -58,3 +74,9 @@ logger = 0
 def setLogger(l):
     global logger
     logger = l
+
+def hexdump(octets):
+    return ' '.join(
+            [ '%s%.2X' % (n%16 == 0 and ('\n%.5d: ' % n) or '', x) 
+              for n,x in zip(range(len(octets)), octs2ints(octets)) ]
+        )
