@@ -1,20 +1,11 @@
 # Initial SNMP engine configuration functions. During further operation,
 # SNMP engine might be configured remotely (through SNMP).
 from pyasn1.compat.octets import null
-from pysnmp.carrier.asynsock import dispatch
-from pysnmp.carrier.asynsock.dgram import udp, udp6
-try:
-    from pysnmp.carrier.asynsock.dgram import unix
-    snmpLocalDomain = unix.snmpLocalDomain
-except ImportError: # UNIX-specific -- may not be always available
-    pass
-from pysnmp.proto import rfc3412
-from pysnmp.entity import engine
+from pysnmp.carrier.asynsock.dgram import udp, udp6, unix
 from pysnmp.proto.secmod.rfc3414.auth import hmacmd5, hmacsha, noauth
 from pysnmp.proto.secmod.rfc3414.priv import des, nopriv
 from pysnmp.proto.secmod.rfc3826.priv import aes
 from pysnmp.proto.secmod.eso.priv import des3, aes192, aes256
-from pysnmp.smi.error import NotWritableError
 from pysnmp import error
 
 # A shortcut to popular constants
@@ -22,6 +13,7 @@ from pysnmp import error
 # Transports
 snmpUDPDomain = udp.snmpUDPDomain
 snmpUDP6Domain = udp6.snmpUDP6Domain
+snmpLocalDomain = unix.snmpLocalDomain
 
 # Auth protocol
 usmHMACMD5AuthProtocol = hmacmd5.HmacMd5.serviceID
@@ -53,71 +45,82 @@ privServices = {
     nopriv.NoPriv.serviceID: nopriv.NoPriv()
     }
     
-def __cookV1SystemInfo(snmpEngine, securityName):
+def __cookV1SystemInfo(snmpEngine, communityIndex):
     snmpEngineID, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineID')
 
     snmpCommunityEntry, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('SNMP-COMMUNITY-MIB', 'snmpCommunityEntry')
-    tblIdx = snmpCommunityEntry.getInstIdFromIndices(securityName)
+    tblIdx = snmpCommunityEntry.getInstIdFromIndices(communityIndex)
     return snmpCommunityEntry, tblIdx, snmpEngineID
     
-def addV1System(snmpEngine, securityName, communityName,
+def addV1System(snmpEngine, communityIndex, communityName,
                 contextEngineId=None, contextName=None,
-                transportTag=None):
+                transportTag=None, securityName=None):
     snmpCommunityEntry, tblIdx, snmpEngineID = __cookV1SystemInfo(
-        snmpEngine, securityName
-        )
+        snmpEngine, communityIndex
+    )
 
     if contextEngineId is None:
         contextEngineId = snmpEngineID.syntax
+    else:
+        contextEngineId = snmpEngineID.syntax.clone(contextEngineId)
+
     if contextName is None:
         contextName = null
 
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
         ((snmpCommunityEntry.name + (8,) + tblIdx, 'destroy'),)
-        )
+    )
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
         ((snmpCommunityEntry.name + (8,) + tblIdx, 'createAndGo'),
+         (snmpCommunityEntry.name + (1,) + tblIdx, communityIndex),
          (snmpCommunityEntry.name + (2,) + tblIdx, communityName),
-         (snmpCommunityEntry.name + (3,) + tblIdx, securityName),
+         (snmpCommunityEntry.name + (3,) + tblIdx, securityName is not None and securityName or communityIndex),
          (snmpCommunityEntry.name + (4,) + tblIdx, contextEngineId),
          (snmpCommunityEntry.name + (5,) + tblIdx, contextName),
          (snmpCommunityEntry.name + (6,) + tblIdx, transportTag),
          (snmpCommunityEntry.name + (7,) + tblIdx, 'nonVolatile'))
-        )
+    )
 
-def delV1System(snmpEngine, securityName):
+def delV1System(snmpEngine, communityIndex):
     snmpCommunityEntry, tblIdx, snmpEngineID = __cookV1SystemInfo(
-        snmpEngine, securityName
-        )
+        snmpEngine, communityIndex
+    )
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
         ((snmpCommunityEntry.name + (8,) + tblIdx, 'destroy'),)
-        )
+    )
 
-def __cookV3UserInfo(snmpEngine, securityName, contextEngineId):
-    if contextEngineId is None:
-        snmpEngineID, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineID')
+def __cookV3UserInfo(snmpEngine, securityName, securityEngineId):
+    snmpEngineID, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineID')
+    if securityEngineId is None:
         snmpEngineID = snmpEngineID.syntax
     else:
-        snmpEngineID = contextEngineId
+        snmpEngineID = snmpEngineID.syntax.clone(securityEngineId)
 
     usmUserEntry, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('SNMP-USER-BASED-SM-MIB', 'usmUserEntry')
     tblIdx1 = usmUserEntry.getInstIdFromIndices(
         snmpEngineID, securityName
-        )
+    )
 
     pysnmpUsmSecretEntry, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('PYSNMP-USM-MIB', 'pysnmpUsmSecretEntry')
     tblIdx2 = pysnmpUsmSecretEntry.getInstIdFromIndices(securityName)
 
     return snmpEngineID, usmUserEntry, tblIdx1, pysnmpUsmSecretEntry, tblIdx2
 
-def addV3User(snmpEngine, securityName,
+def addV3User(snmpEngine, userName,
               authProtocol=usmNoAuthProtocol, authKey=None,
               privProtocol=usmNoPrivProtocol, privKey=None,
+              securityEngineId=None,
+              securityName=None,
+              # deprecated parameters follow
               contextEngineId=None):
+    if securityName is None:
+        securityName = userName
+    if securityEngineId is None:  # backward compatibility
+        securityEngineId = contextEngineId
     ( snmpEngineID, usmUserEntry, tblIdx1,
       pysnmpUsmSecretEntry, tblIdx2 ) = __cookV3UserInfo(
-        snmpEngine, securityName, contextEngineId
-        )
+        snmpEngine, userName, securityEngineId
+    )
 
     # Load augmenting table before creating new row in base one
     pysnmpUsmKeyEntry, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('PYSNMP-USM-MIB', 'pysnmpUsmKeyEntry')
@@ -127,37 +130,38 @@ def addV3User(snmpEngine, securityName,
 
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
         ((usmUserEntry.name + (13,) + tblIdx1, 'destroy'),)
-        )
+    )
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
         ((usmUserEntry.name + (13,) + tblIdx1, 'createAndGo'),
+         (usmUserEntry.name + (2,) + tblIdx1, userName),
          (usmUserEntry.name + (3,) + tblIdx1, securityName),
          (usmUserEntry.name + (4,) + tblIdx1, zeroDotZero.name),
          (usmUserEntry.name + (5,) + tblIdx1, authProtocol),
          (usmUserEntry.name + (8,) + tblIdx1, privProtocol))
-        )
+    )
 
     # Localize keys
     if authProtocol in authServices:
         hashedAuthPassphrase = authServices[authProtocol].hashPassphrase(
             authKey and authKey or null
-            )
+        )
         localAuthKey = authServices[authProtocol].localizeKey(
             hashedAuthPassphrase, snmpEngineID
-            )
+        )
     else:
         raise error.PySnmpError('Unknown auth protocol %s' % (authProtocol,))
 
     if privProtocol in privServices:
         hashedPrivPassphrase = privServices[privProtocol].hashPassphrase(
             authProtocol, privKey and privKey or null
-            )
+        )
         localPrivKey = privServices[privProtocol].localizeKey(
             authProtocol, hashedPrivPassphrase, snmpEngineID
-            )
+        )
     else:
         raise error.PySnmpError(
             'Unknown priv protocol %s' % (privProtocol,)
-            )
+        )
 
     # Commit localized keys
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
@@ -178,17 +182,39 @@ def addV3User(snmpEngine, securityName,
          (pysnmpUsmSecretEntry.name + (3,) + tblIdx2, privKey),)
         )
 
-def delV3User(snmpEngine, securityName, contextEngineId=None):
+def delV3User(snmpEngine,
+              userName,
+              securityEngineId=None,
+              # deprecated parameters follow
+              contextEngineId=None):
+    if securityEngineId is None:  # backward compatibility
+        securityEngineId = contextEngineId
     ( snmpEngineID, usmUserEntry, tblIdx1,
       pysnmpUsmSecretEntry, tblIdx2 ) = __cookV3UserInfo(
-        snmpEngine, securityName, contextEngineId
-        )
+        snmpEngine, userName, securityEngineId
+    )
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
         ((usmUserEntry.name + (13,) + tblIdx1, 'destroy'),)
-        )
+    )
     snmpEngine.msgAndPduDsp.mibInstrumController.writeVars(
         ((pysnmpUsmSecretEntry.name + (4,) + tblIdx2, 'destroy'),)
+    )
+
+    # Drop all derived rows
+    varBinds = initialVarBinds = (
+        (usmUserEntry.name + (1,), None),  # usmUserEngineID
+        (usmUserEntry.name + (2,), None),  # usmUserName
+        (usmUserEntry.name + (4,), None)   # usmUserCloneFrom
+    )
+    while varBinds:
+        varBinds = snmpEngine.msgAndPduDsp.mibInstrumController.readNextVars(
+            varBinds
         )
+        if varBinds[0][0][:len(initialVarBinds[0][0])]!=initialVarBinds[0][0]:
+            break
+        elif varBinds[2][1] == tblIdx1:  # cloned from this entry
+            delV3User(snmpEngine, varBinds[1][1], varBinds[0][1])
+            varBinds = initialVarBinds
 
 def __cookTargetParamsInfo(snmpEngine, name):
     snmpTargetParamsEntry, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('SNMP-TARGET-MIB', 'snmpTargetParamsEntry')
@@ -269,8 +295,7 @@ def addTargetAddr(
         (snmpTargetAddrEntry.name + (3,) + tblIdx, transportAddress),
         (snmpTargetAddrEntry.name + (4,) + tblIdx, timeout),
         (snmpTargetAddrEntry.name + (5,) + tblIdx, retryCount),
-         # XXX
-        (snmpTargetAddrEntry.name + (6,) + tblIdx,'%s%s%s' % (addrName, tagList and ' ' or '', tagList)),
+        (snmpTargetAddrEntry.name + (6,) + tblIdx, tagList),
         (snmpTargetAddrEntry.name + (7,) + tblIdx, params),)
         )
 
@@ -282,20 +307,36 @@ def delTargetAddr(snmpEngine, addrName):
         ((snmpTargetAddrEntry.name + (9,) + tblIdx, 'destroy'),)
         )
 
-def addSocketTransport(snmpEngine, transportDomain, transport):
-    """Add transport object to socket dispatcher of snmpEngine"""
-    if not snmpEngine.transportDispatcher:
-        snmpEngine.registerTransportDispatcher(dispatch.AsynsockDispatcher())
+def addTransport(snmpEngine, transportDomain, transport):
+    if snmpEngine.transportDispatcher:
+        if not transport.isCompatibleWithDispatcher(snmpEngine.transportDispatcher): 
+            raise error.PySnmpError('Transport %r is not compatible with dispatcher %r' % (transport, snmpEngine.transportDispatcher))
+    else:
+        snmpEngine.registerTransportDispatcher(
+            transport.protoTransportDispatcher()
+        )
+        # here we note that we have created transportDispatcher automatically
+        snmpEngine.cache['automaticTransportDispatcher'] = 0
     snmpEngine.transportDispatcher.registerTransport(
         transportDomain, transport
-        )
+    )
+    if 'automaticTransportDispatcher' in snmpEngine.cache:
+        snmpEngine.cache['automaticTransportDispatcher'] += 1
 
-def delSocketTransport(snmpEngine, transportDomain):
-    """Unregister transport object at socket dispatcher of snmpEngine"""
+def delTransport(snmpEngine, transportDomain):
     if not snmpEngine.transportDispatcher:
         return
     snmpEngine.transportDispatcher.unregisterTransport(transportDomain)
-    snmpEngine.unregisterTransportDispatcher()
+    # automatically shutdown automatically created transportDispatcher
+    if 'automaticTransportDispatcher' in snmpEngine.cache:
+        snmpEngine.cache['automaticTransportDispatcher'] -= 1
+        if not snmpEngine.cache['automaticTransportDispatcher']:
+            snmpEngine.transportDispatcher.closeDispatcher()
+            snmpEngine.unregisterTransportDispatcher()
+            del snmpEngine.cache['automaticTransportDispatcher']
+
+addSocketTransport = addTransport
+delSocketTransport = delTransport
 
 # VACM shortcuts
 
