@@ -1,19 +1,14 @@
-import time
-from pysnmp.proto import rfc1157, rfc1905, api, errind
+import sys
 from pysnmp.entity.rfc3413 import config
+from pysnmp.proto import rfc1905, errind
+from pysnmp.proto.api import v2c
 from pysnmp.proto.proxy import rfc2576
 from pysnmp import error, nextid, debug
+from pysnmp.proto.error import StatusInformation
 from pyasn1.type import univ
 
 getNextHandle = nextid.Integer(0x7fffffff)
                              
-def getVersionSpecifics(snmpVersion):
-    if snmpVersion == 0:
-        pduVersion = 0
-    else:
-        pduVersion = 1
-    return pduVersion, api.protoModules[pduVersion]
-
 __null = univ.Null('')
     
 def getNextVarBinds(origVarBinds, varBinds):
@@ -22,7 +17,9 @@ def getNextVarBinds(origVarBinds, varBinds):
     rspVarBinds = []
     while idx:
         idx = idx - 1
-        if isinstance(varBinds[idx][1], univ.Null):
+        if varBinds[idx][1].tagSet in (rfc1905.NoSuchObject.tagSet,
+                                       rfc1905.NoSuchInstance.tagSet,
+                                       rfc1905.EndOfMibView.tagSet):
             nonNulls = nonNulls - 1
         elif origVarBinds[idx][0].asTuple() >= varBinds[idx][0].asTuple():
             errorIndication = errind.oidNotIncreasing
@@ -57,6 +54,9 @@ class CommandGeneratorBase:
         ):
         (cbFun, cbCtx) = cbInfo
         # 3.1.1
+        if sendPduHandle not in self.__pendingReqs:
+            raise error.PySnmpError('Missing sendPduHandle %s' % sendPduHandle)
+
         ( origTransportDomain,
           origTransportAddress,
           origMessageProcessingModel,
@@ -70,8 +70,8 @@ class CommandGeneratorBase:
           origTimeout,
           origRetryCount,
           origRetries,
-          origSendRequestHandle
-          ) = self.__pendingReqs[sendPduHandle]
+          origSendRequestHandle ) = self.__pendingReqs[sendPduHandle]
+  
         del self.__pendingReqs[sendPduHandle]
 
         snmpEngine.transportDispatcher.jobFinished(id(self))
@@ -79,30 +79,43 @@ class CommandGeneratorBase:
         # 3.1.3
         if statusInformation:
             debug.logger & debug.flagApp and debug.logger('processResponsePdu: sendPduHandle %s, statusInformation %s' % (sendPduHandle, statusInformation))
-            if origRetries == origRetryCount:
+            errorIndication = statusInformation['errorIndication']
+            # SNMP engine discovery will take extra retries, allow that
+            if errorIndication in (errind.notInTimeWindow,
+                                   errind.unknownEngineID) and \
+                                   origRetries == origRetryCount + 2 or \
+               errorIndication not in (errind.notInTimeWindow,
+                                       errind.unknownEngineID) and \
+                                   origRetries == origRetryCount:
                 debug.logger & debug.flagApp and debug.logger('processResponsePdu: sendPduHandle %s, retry count %d exceeded' % (sendPduHandle, origRetries))
                 cbFun(origSendRequestHandle,
                       statusInformation['errorIndication'], 0, 0, (),
                       cbCtx)
                 return
-            self._sendPdu(
-                snmpEngine,
-                origTransportDomain,
-                origTransportAddress,
-                origMessageProcessingModel,
-                origSecurityModel,
-                origSecurityName,
-                origSecurityLevel,
-                origContextEngineId,
-                origContextName,
-                origPduVersion,
-                origPdu,
-                origTimeout,
-                origRetryCount,
-                origRetries + 1,
-                origSendRequestHandle,
-                (self.processResponsePdu, (cbFun, cbCtx))
+            try:
+                self._sendPdu(
+                    snmpEngine,
+                    origTransportDomain,
+                    origTransportAddress,
+                    origMessageProcessingModel,
+                    origSecurityModel,
+                    origSecurityName,
+                    origSecurityLevel,
+                    origContextEngineId,
+                    origContextName,
+                    origPdu,
+                    origTimeout,
+                    origRetryCount,
+                    origRetries + 1,
+                    origSendRequestHandle,
+                    (self.processResponsePdu, (cbFun, cbCtx))
                 )
+            except StatusInformation:
+                statusInformation = sys.exc_info()[1]
+                debug.logger & debug.flagApp and debug.logger('processResponsePdu: origSendRequestHandle %s, _sendPdu() failed with %r' % (sendPduHandle, statusInformation))
+                cbFun(origSendRequestHandle,
+                      statusInformation['errorIndication'], 0, 0, (),
+                      cbCtx)
             return
 
         if origMessageProcessingModel != messageProcessingModel or \
@@ -115,19 +128,16 @@ class CommandGeneratorBase:
             cbFun(origSendRequestHandle, 'badResponse', 0, 0, (), cbCtx)
             return
 
-        pMod = api.protoModules[pduVersion]
-        
+        # User-side API assumes SMIv2
+        if messageProcessingModel == 0:
+            PDU = rfc2576.v1ToV2(PDU, origPdu)
+ 
         # 3.1.2
-        if pMod.apiPDU.getRequestID(PDU) != pMod.apiPDU.getRequestID(origPdu):
+        if v2c.apiPDU.getRequestID(PDU) != v2c.apiPDU.getRequestID(origPdu):
             debug.logger & debug.flagApp and debug.logger('processResponsePdu: sendPduHandle %s, request-id/response-id mismatch' % sendPduHandle)
             cbFun(origSendRequestHandle, 'badResponse', 0, 0, (), cbCtx)
             return
 
-        # User-side API assumes SMIv2
-        if messageProcessingModel == 0:
-            PDU = rfc2576.v1ToV2(PDU, origPdu)
-            pMod = api.protoModules[api.protoVersion2c]
-        
         self._handleResponse(
             snmpEngine,
             origTransportDomain,
@@ -138,15 +148,13 @@ class CommandGeneratorBase:
             origSecurityLevel,
             origContextEngineId,
             origContextName,
-            origPduVersion,
             origPdu,
             origTimeout,
             origRetryCount,
-            pMod,
             PDU,
             origSendRequestHandle,
             (cbFun, cbCtx),
-            )
+        )
 
     def sendReq(
         self,
@@ -171,7 +179,6 @@ class CommandGeneratorBase:
         securityLevel,
         contextEngineId,
         contextName,
-        pduVersion,
         reqPDU,
         timeout,
         retryCount,
@@ -192,6 +199,15 @@ class CommandGeneratorBase:
             contextEngineId = self.__SnmpEngineID(contextEngineId)
         contextName = self.__SnmpAdminString(contextName)
 
+        origPDU = reqPDU
+
+        # User-side API assumes SMIv2
+        if messageProcessingModel == 0:
+            reqPDU = rfc2576.v2ToV1(reqPDU)
+            pduVersion = 0
+        else:
+            pduVersion = 1
+ 
         # 3.1
         sendPduHandle = snmpEngine.msgAndPduDsp.sendPdu(
             snmpEngine,
@@ -205,16 +221,14 @@ class CommandGeneratorBase:
             contextName,
             pduVersion,
             reqPDU,
-            1,                          # expectResponse
+            1, # expectResponse
             timeoutInTicks,
             processResponsePdu,
             cbCtx
-            )
+        )
 
         snmpEngine.transportDispatcher.jobStarted(id(self))
 
-        debug.logger & debug.flagApp and debug.logger('_sendPdu: sendPduHandle %s, timeout %d, retry %d of %d' % (sendPduHandle, timeout, retries, retryCount))
-
         self.__pendingReqs[sendPduHandle] = (
             transportDomain,
             transportAddress,
@@ -225,12 +239,40 @@ class CommandGeneratorBase:
             contextEngineId,
             contextName,
             pduVersion,
-            reqPDU,
+            origPDU,
             timeout,
             retryCount,
             retries,
-            sendRequestHandle,
-            )
+            sendRequestHandle
+        )
+ 
+        debug.logger & debug.flagApp and debug.logger('_sendPdu: sendPduHandle %s, timeout %d*10 ms/%d ticks, retry %d of %d' % (sendPduHandle, timeout, timeoutInTicks, retries, retryCount))
+
+    def _handleResponse(
+        self,
+        snmpEngine,
+        transportDomain,
+        transportAddress,
+        messageProcessingModel,
+        securityModel,
+        securityName,
+        securityLevel,
+        contextEngineId,
+        contextName,
+        PDU,
+        timeout,
+        retryCount,
+        rspPDU,
+        sendRequestHandle,
+        cbInfo
+        ):
+        (cbFun, cbCtx) = cbInfo        
+        cbFun(sendRequestHandle,
+              None,
+              v2c.apiPDU.getErrorStatus(rspPDU),
+              v2c.apiPDU.getErrorIndex(rspPDU, muteErrors=True),
+              v2c.apiPDU.getVarBinds(rspPDU),
+              cbCtx)
 
 class GetCommandGenerator(CommandGeneratorBase):
     def sendReq(
@@ -252,64 +294,39 @@ class GetCommandGenerator(CommandGeneratorBase):
           securityName,
           securityLevel ) = config.getTargetInfo(snmpEngine, addrName)
 
-        pduVersion, pMod = getVersionSpecifics(messageProcessingModel)
+        reqPDU = v2c.GetRequestPDU()
+        v2c.apiPDU.setDefaults(reqPDU)
         
-        reqPDU = pMod.GetRequestPDU()
-        pMod.apiPDU.setDefaults(reqPDU)
-        
-        pMod.apiPDU.setVarBinds(reqPDU, varBinds)
+        v2c.apiPDU.setVarBinds(reqPDU, varBinds)
 
         requestHandle = getNextHandle()
-        
-        self._sendPdu(
-            snmpEngine,
-            transportDomain,
-            transportAddress,
-            messageProcessingModel,
-            securityModel,
-            securityName,
-            securityLevel,
-            contextEngineId,
-            contextName,
-            pduVersion,
-            reqPDU,
-            timeout,
-            retryCount,
-            0,
-            requestHandle,
-            (self.processResponsePdu, (cbFun, cbCtx))            
+
+        try:        
+            self._sendPdu(
+                snmpEngine,
+                transportDomain,
+                transportAddress,
+                messageProcessingModel,
+                securityModel,
+                securityName,
+                securityLevel,
+                contextEngineId,
+                contextName,
+                reqPDU,
+                timeout,
+                retryCount,
+                0, # retries
+                requestHandle,
+                (self.processResponsePdu, (cbFun, cbCtx))            
             )
+        except StatusInformation:
+            statusInformation = sys.exc_info()[1]
+            debug.logger & debug.flagApp and debug.logger('sendReq: sendPduHandle %s: _sendPdu() failed with %r' % (requestHandle, statusInformation))
+            cbFun(requestHandle, statusInformation['errorIndication'],
+                  0, 0, (), cbCtx)
 
         return requestHandle
     
-    def _handleResponse(
-        self,
-        snmpEngine,
-        transportDomain,
-        transportAddress,
-        messageProcessingModel,
-        securityModel,
-        securityName,
-        securityLevel,
-        contextEngineId,
-        contextName,
-        pduVersion,
-        PDU,
-        timeout,
-        retryCount,
-        pMod,
-        rspPDU,
-        sendRequestHandle,
-        cbInfo
-        ):
-        (cbFun, cbCtx) = cbInfo        
-        cbFun(sendRequestHandle,
-              None,
-              pMod.apiPDU.getErrorStatus(rspPDU),
-              pMod.apiPDU.getErrorIndex(rspPDU),
-              pMod.apiPDU.getVarBinds(rspPDU),
-              cbCtx)
-
 class SetCommandGenerator(CommandGeneratorBase):
     def sendReq(
         self,
@@ -330,70 +347,40 @@ class SetCommandGenerator(CommandGeneratorBase):
           securityName,
           securityLevel ) = config.getTargetInfo(snmpEngine, addrName)
 
-        pduVersion, pMod = getVersionSpecifics(messageProcessingModel)
-        
-        reqPDU = pMod.SetRequestPDU()
-        pMod.apiPDU.setDefaults(reqPDU)
+        reqPDU = v2c.SetRequestPDU()
+        v2c.apiPDU.setDefaults(reqPDU)
 
-        pMod.apiPDU.setVarBinds(reqPDU, varBinds)
-
-        # User-side API assumes SMIv2
-        if messageProcessingModel == 0:
-            reqPDU = rfc2576.v2ToV1(reqPDU)
-            pMod = api.protoModules[api.protoVersion1]
+        v2c.apiPDU.setVarBinds(reqPDU, varBinds)
 
         requestHandle = getNextHandle()        
-        
-        self._sendPdu(
-            snmpEngine,
-            transportDomain,
-            transportAddress,
-            messageProcessingModel,
-            securityModel,
-            securityName,
-            securityLevel,
-            contextEngineId,
-            contextName,
-            pduVersion,
-            reqPDU,
-            timeout,
-            retryCount,
-            0,
-            requestHandle,
-            (self.processResponsePdu, (cbFun, cbCtx))            
+
+        try:        
+            self._sendPdu(
+                snmpEngine,
+                transportDomain,
+                transportAddress,
+                messageProcessingModel,
+                securityModel,
+                securityName,
+                securityLevel,
+                contextEngineId,
+                contextName,
+                reqPDU,
+                timeout,
+                retryCount,
+                0, # retries
+                requestHandle,
+                (self.processResponsePdu, (cbFun, cbCtx))            
             )
+        except StatusInformation:
+            statusInformation = sys.exc_info()[1]
+            debug.logger & debug.flagApp and debug.logger('sendReq: sendPduHandle %s: _sendPdu() failed with %r' % (requestHandle, statusInformation))
+            cbFun(requestHandle, statusInformation['errorIndication'],
+                  0, 0, (), cbCtx)
 
         return requestHandle
 
-    def _handleResponse(
-        self,
-        snmpEngine,
-        transportDomain,
-        transportAddress,
-        messageProcessingModel,
-        securityModel,
-        securityName,
-        securityLevel,
-        contextEngineId,
-        contextName,
-        pduVersion,
-        PDU,
-        timeout,
-        retryCount,
-        pMod,
-        rspPDU,
-        sendRequestHandle,
-        cbInfo
-        ):
-        (cbFun, cbCtx) = cbInfo        
-        cbFun(sendRequestHandle,
-              None,
-              pMod.apiPDU.getErrorStatus(rspPDU),
-              pMod.apiPDU.getErrorIndex(rspPDU),
-              pMod.apiPDU.getVarBinds(rspPDU),
-              cbCtx)
-
-class NextCommandGenerator(CommandGeneratorBase):
+class NextCommandGeneratorSingleRun(CommandGeneratorBase):
     def sendReq(
         self,
         snmpEngine,
@@ -413,36 +400,40 @@ class NextCommandGenerator(CommandGeneratorBase):
           securityName,
           securityLevel ) = config.getTargetInfo(snmpEngine, addrName)
 
-        pduVersion, pMod = getVersionSpecifics(messageProcessingModel)
+        reqPDU = v2c.GetNextRequestPDU()
+        v2c.apiPDU.setDefaults(reqPDU)
         
-        reqPDU = pMod.GetNextRequestPDU()
-        pMod.apiPDU.setDefaults(reqPDU)
-        
-        pMod.apiPDU.setVarBinds(reqPDU, varBinds)
+        v2c.apiPDU.setVarBinds(reqPDU, varBinds)
 
         requestHandle = getNextHandle()        
-        
-        self._sendPdu(
-            snmpEngine,
-            transportDomain,
-            transportAddress,
-            messageProcessingModel,
-            securityModel,
-            securityName,
-            securityLevel,
-            contextEngineId,
-            contextName,
-            pduVersion,
-            reqPDU,
-            timeout,
-            retryCount,
-            0,
-            requestHandle,
-            (self.processResponsePdu, (cbFun, cbCtx))            
+
+        try:        
+            self._sendPdu(
+                snmpEngine,
+                transportDomain,
+                transportAddress,
+                messageProcessingModel,
+                securityModel,
+                securityName,
+                securityLevel,
+                contextEngineId,
+                contextName,
+                reqPDU,
+                timeout,
+                retryCount,
+                0, # retries
+                requestHandle,
+                (self.processResponsePdu, (cbFun, cbCtx))            
             )
+        except StatusInformation:
+            statusInformation = sys.exc_info()[1]
+            debug.logger & debug.flagApp and debug.logger('sendReq: sendPduHandle %s: _sendPdu() failed with %r' % (requestHandle, statusInformation))
+            cbFun(requestHandle, statusInformation['errorIndication'],
+                  0, 0, (), cbCtx)
 
         return requestHandle
-    
+ 
+class NextCommandGenerator(NextCommandGeneratorSingleRun):
     def _handleResponse(
         self,
         snmpEngine,
@@ -454,30 +445,29 @@ class NextCommandGenerator(CommandGeneratorBase):
         securityLevel,
         contextEngineId,
         contextName,
-        pduVersion,
         PDU,
         timeout,
         retryCount,
-        pMod,
         rspPDU,
         sendRequestHandle,
         cbInfo
         ):
         (cbFun, cbCtx) = cbInfo
-        varBindTable = pMod.apiPDU.getVarBindTable(PDU, rspPDU)
 
-        if pMod.apiPDU.getErrorStatus(rspPDU):
+        varBindTable = v2c.apiPDU.getVarBindTable(PDU, rspPDU)
+
+        if v2c.apiPDU.getErrorStatus(rspPDU):
             errorIndication, varBinds = None, ()
         elif not varBindTable:
             errorIndication, varBinds = errind.emptyResponse, ()
         else:
             errorIndication, varBinds = getNextVarBinds(
-                pMod.apiPDU.getVarBinds(PDU), varBindTable[-1]
-                )
+                v2c.apiPDU.getVarBinds(PDU), varBindTable[-1]
+            )
         
         if not cbFun(sendRequestHandle, errorIndication,
-                     pMod.apiPDU.getErrorStatus(rspPDU),
-                     pMod.apiPDU.getErrorIndex(rspPDU),
+                     v2c.apiPDU.getErrorStatus(rspPDU),
+                     v2c.apiPDU.getErrorIndex(rspPDU, muteErrors=True),
                      varBindTable, cbCtx):
             debug.logger & debug.flagApp and debug.logger('_handleResponse: sendRequestHandle %s, app says to stop walking' % sendRequestHandle)
             return  # app says enough
@@ -485,29 +475,36 @@ class NextCommandGenerator(CommandGeneratorBase):
         if not varBinds:
             return # no more objects available
     
-        pMod.apiPDU.setRequestID(PDU, pMod.getNextRequestID())
-        pMod.apiPDU.setVarBinds(PDU, varBinds)
+        v2c.apiPDU.setRequestID(PDU, v2c.getNextRequestID())
+        v2c.apiPDU.setVarBinds(PDU, varBinds)
 
-        self._sendPdu(
-            snmpEngine,
-            transportDomain,
-            transportAddress,
-            messageProcessingModel,
-            securityModel,
-            securityName,
-            securityLevel,
-            contextEngineId,
-            contextName,
-            pduVersion,
-            PDU,
-            timeout,
-            retryCount,
-            0,
-            getNextHandle(),
-            (self.processResponsePdu, (cbFun, cbCtx))            
+        sendRequestHandle = getNextHandle()
+
+        try:
+            self._sendPdu(
+                snmpEngine,
+                transportDomain,
+                transportAddress,
+                messageProcessingModel,
+                securityModel,
+                securityName,
+                securityLevel,
+                contextEngineId,
+                contextName,
+                PDU,
+                timeout,
+                retryCount,
+                0, # retries
+                sendRequestHandle,
+                (self.processResponsePdu, (cbFun, cbCtx))            
             )
-
-class BulkCommandGenerator(CommandGeneratorBase):
+        except StatusInformation:
+            statusInformation = sys.exc_info()[1]
+            debug.logger & debug.flagApp and debug.logger('sendReq: sendPduHandle %s: _sendPdu() failed with %r' % (sendRequestHandle, statusInformation))
+            cbFun(sendRequestHandle, statusInformation['errorIndication'],
+                  0, 0, (), cbCtx)
+ 
+class BulkCommandGeneratorSingleRun(CommandGeneratorBase):
     def sendReq(
         self,
         snmpEngine,
@@ -529,41 +526,43 @@ class BulkCommandGenerator(CommandGeneratorBase):
           securityName,
           securityLevel ) = config.getTargetInfo(snmpEngine, addrName)
 
-        pduVersion, pMod = getVersionSpecifics(messageProcessingModel)
-       
-        if not hasattr(pMod, 'GetBulkRequestPDU'):
-            raise error.PySnmpError('BULK PDU not implemented at %s' % pMod)
-        reqPDU = pMod.GetBulkRequestPDU()
-        pMod.apiBulkPDU.setDefaults(reqPDU)
+        reqPDU = v2c.GetBulkRequestPDU()
+        v2c.apiBulkPDU.setDefaults(reqPDU)
         
-        pMod.apiBulkPDU.setNonRepeaters(reqPDU, nonRepeaters)
-        pMod.apiBulkPDU.setMaxRepetitions(reqPDU, maxRepetitions)
+        v2c.apiBulkPDU.setNonRepeaters(reqPDU, nonRepeaters)
+        v2c.apiBulkPDU.setMaxRepetitions(reqPDU, maxRepetitions)
 
-        pMod.apiBulkPDU.setVarBinds(reqPDU, varBinds)
+        v2c.apiBulkPDU.setVarBinds(reqPDU, varBinds)
 
         requestHandle = getNextHandle()        
-        
-        self._sendPdu(
-            snmpEngine,
-            transportDomain,
-            transportAddress,
-            messageProcessingModel,
-            securityModel,
-            securityName,
-            securityLevel,
-            contextEngineId,
-            contextName,
-            pduVersion,
-            reqPDU,
-            timeout,
-            retryCount,
-            0,
-            requestHandle,
-            (self.processResponsePdu, (cbFun, cbCtx))            
+
+        try:        
+            self._sendPdu(
+                snmpEngine,
+                transportDomain,
+                transportAddress,
+                messageProcessingModel,
+                securityModel,
+                securityName,
+                securityLevel,
+                contextEngineId,
+                contextName,
+                reqPDU,
+                timeout,
+                retryCount,
+                0, # retries
+                requestHandle,
+                (self.processResponsePdu, (cbFun, cbCtx))            
             )
+        except StatusInformation:
+            statusInformation = sys.exc_info()[1]
+            debug.logger & debug.flagApp and debug.logger('sendReq: sendPduHandle %s: _sendPdu() failed with %r' % (requestHandle, statusInformation))
+            cbFun(requestHandle, statusInformation['errorIndication'],
+                  0, 0, (), cbCtx)
 
         return requestHandle
-    
+ 
+class BulkCommandGenerator(BulkCommandGeneratorSingleRun):
     def _handleResponse(
         self,
         snmpEngine,
@@ -575,30 +574,28 @@ class BulkCommandGenerator(CommandGeneratorBase):
         securityLevel,
         contextEngineId,
         contextName,
-        pduVersion,
         PDU,
         timeout,
         retryCount,
-        pMod,
         rspPDU,
         sendRequestHandle,
         cbInfo
         ):
         (cbFun, cbCtx) = cbInfo        
-        varBindTable = pMod.apiBulkPDU.getVarBindTable(PDU, rspPDU)
+        varBindTable = v2c.apiBulkPDU.getVarBindTable(PDU, rspPDU)
 
-        if pMod.apiBulkPDU.getErrorStatus(rspPDU):
+        if v2c.apiBulkPDU.getErrorStatus(rspPDU):
             errorIndication, varBinds = None, ()
         elif not varBindTable:
             errorIndication, varBinds = errind.emptyResponse, ()
         else:
             errorIndication, varBinds = getNextVarBinds(
-                pMod.apiBulkPDU.getVarBinds(PDU), varBindTable[-1]
-                )
+                v2c.apiBulkPDU.getVarBinds(PDU), varBindTable[-1]
+            )
 
         if not cbFun(sendRequestHandle, errorIndication,
-                     pMod.apiBulkPDU.getErrorStatus(rspPDU),
-                     pMod.apiBulkPDU.getErrorIndex(rspPDU),
+                     v2c.apiBulkPDU.getErrorStatus(rspPDU),
+                     v2c.apiBulkPDU.getErrorIndex(rspPDU, muteErrors=True),
                      varBindTable, cbCtx):
             debug.logger & debug.flagApp and debug.logger('_handleResponse: sendRequestHandle %s, app says to stop walking' % sendRequestHandle)
             return # app says enough
@@ -606,24 +603,31 @@ class BulkCommandGenerator(CommandGeneratorBase):
         if not varBinds:
             return # no more objects available
     
-        pMod.apiBulkPDU.setRequestID(PDU, pMod.getNextRequestID())
-        pMod.apiBulkPDU.setVarBinds(PDU, varBinds)
-        
-        self._sendPdu(
-            snmpEngine,
-            transportDomain,
-            transportAddress,
-            messageProcessingModel,
-            securityModel,
-            securityName,
-            securityLevel,
-            contextEngineId,
-            contextName,
-            pduVersion,
-            PDU,
-            timeout,
-            retryCount,
-            0,
-            getNextHandle(),
-            (self.processResponsePdu, (cbFun, cbCtx))            
+        v2c.apiBulkPDU.setRequestID(PDU, v2c.getNextRequestID())
+        v2c.apiBulkPDU.setVarBinds(PDU, varBinds)
+
+        sendRequestHandle = getNextHandle()
+
+        try:        
+            self._sendPdu(
+                snmpEngine,
+                transportDomain,
+                transportAddress,
+                messageProcessingModel,
+                securityModel,
+                securityName,
+                securityLevel,
+                contextEngineId,
+                contextName,
+                PDU,
+                timeout,
+                retryCount,
+                0, # retries
+                sendRequestHandle,
+                (self.processResponsePdu, (cbFun, cbCtx))            
             )
+        except StatusInformation:
+            statusInformation = sys.exc_info()[1]
+            debug.logger & debug.flagApp and debug.logger('sendReq: sendPduHandle %s: _sendPdu() failed with %r' % (sendRequestHandle, statusInformation))
+            cbFun(sendRequestHandle, statusInformation['errorIndication'],
+                  0, 0, (), cbCtx)
