@@ -1,7 +1,11 @@
-from twisted.internet import reactor, defer
+from twisted.internet import defer
 from pysnmp.entity.rfc3413 import ntforg
 from pyasn1.compat.octets import null
 
+def _cbFun(sendRequestHandle, errorIndication,
+           errorStatus, errorIndex, varBinds, cbCtx):
+    cbCtx.callback((errorIndication, errorStatus, errorIndex, varBinds))
+
 class NotificationOriginator(ntforg.NotificationOriginator):
     def sendNotification(
         self,
@@ -18,16 +22,8 @@ class NotificationOriginator(ntforg.NotificationOriginator):
             notificationTarget,
             notificationName,
             additionalVarBinds,
-            None,
+            _cbFun,
             df,
             contextName=null
             )
         return df
-
-    def _handleResponse(
-        self,
-        sendRequestHandle,
-        errorIndication,
-        cbFun,
-        cbCtx):
-        cbCtx.callback((sendRequestHandle, errorIndication))
