@@ -56,7 +56,7 @@ class MsgAndPduDispatcher:
             k = (contextEngineId, pduType)
             if k in self.__appsRegistration:
                 raise error.ProtocolError(
-                    'Duplicate registration %s/%s' % (contextEngineId, pduType)
+                    'Duplicate registration %r/%s' % (contextEngineId, pduType)
                     )
 
             # 4.3.4
@@ -131,9 +131,9 @@ class MsgAndPduDispatcher:
                 cbFun=cbFun,
                 cbCtx=cbCtx
                 )
-            debug.logger & debug.flagDsp and debug.logger('sendPdu: current time in ticks %d' % (snmpEngine.transportDispatcher.getTimerTicks(),))
+            debug.logger & debug.flagDsp and debug.logger('sendPdu: current time %d ticks, one tick is %s seconds' % (snmpEngine.transportDispatcher.getTimerTicks(), snmpEngine.transportDispatcher.getTimerResolution()))
 
-        debug.logger & debug.flagDsp and debug.logger('sendPdu: new sendPduHandle %s, timeout %s, cbFun %s' % (sendPduHandle, timeout, cbFun))
+        debug.logger & debug.flagDsp and debug.logger('sendPdu: new sendPduHandle %s, timeout %s ticks, cbFun %s' % (sendPduHandle, timeout, cbFun))
 
         # 4.1.1.4 & 4.1.1.5
         try:
@@ -156,16 +156,26 @@ class MsgAndPduDispatcher:
                 )
             debug.logger & debug.flagDsp and debug.logger('sendPdu: MP succeeded')
         except error.StatusInformation:
+            if expectResponse:
+                self.__cache.pop(sendPduHandle)
 # XXX is it still needed here?
 #            self.releaseStateInformation(snmpEngine, sendPduHandle, messageProcessingModel)
             raise
 
         # 4.1.1.6
         if snmpEngine.transportDispatcher is None:
+            if expectResponse:
+                self.__cache.pop(sendPduHandle)
             raise error.PySnmpError('Transport dispatcher not set')
-        snmpEngine.transportDispatcher.sendMessage(
-            outgoingMessage, destTransportDomain, destTransportAddress
+
+        try:
+            snmpEngine.transportDispatcher.sendMessage(
+                outgoingMessage, destTransportDomain, destTransportAddress
             )
+        except PySnmpError:
+            if expectResponse:
+                self.__cache.pop(sendPduHandle)
+            raise
         
         # Update cache with orignal req params (used for retrying)
         if expectResponse:
@@ -180,7 +190,7 @@ class MsgAndPduDispatcher:
                 contextName=contextName,
                 pduVersion=pduVersion,
                 PDU=PDU
-                )
+            )
 
         return sendPduHandle
 
@@ -241,8 +251,8 @@ class MsgAndPduDispatcher:
                len(outgoingMessage) > snmpEngineMaxMessageSize.syntax:
             snmpSilentDrops, = self.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpSilentDrops')
             snmpSilentDrops.syntax = snmpSilentDrops.syntax + 1
-            raise error.MessageTooBigError()
-        
+            raise error.StatusInformation(errorIndication=errind.tooBig)
+
         # 4.1.2.4
         snmpEngine.transportDispatcher.sendMessage(
             outgoingMessage,
@@ -269,9 +279,9 @@ class MsgAndPduDispatcher:
         try:
             restOfWholeMsg = null # XXX fix decoder non-recursive return
             msgVersion = verdec.decodeMessageVersion(wholeMsg)
-        except PySnmpError:
-            snmpInAsn1ParseErrs, = self.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInAsn1ParseErrs')
-            snmpInAsn1ParseErrs.syntax = snmpInAsn1ParseErrs.syntax + 1
+        except error.ProtocolError:
+            snmpInASNParseErrs, = self.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInASNParseErrs')
+            snmpInASNParseErrs.syntax = snmpInASNParseErrs.syntax + 1
             return null  # n.b the whole buffer gets dropped
 
         debug.logger & debug.flagDsp and debug.logger('receiveMessage: msgVersion %s, msg decoded' % msgVersion)
@@ -314,7 +324,7 @@ class MsgAndPduDispatcher:
             if 'sendPduHandle' in statusInformation:
                 # Dropped REPORT -- re-run pending reqs queue as some
                 # of them may be waiting for this REPORT
-                debug.logger & debug.flagDsp and debug.logger('receiveMessage: MP failed, statusInformation %s' % statusInformation)
+                debug.logger & debug.flagDsp and debug.logger('receiveMessage: MP failed, statusInformation %s, forcing a retry' % statusInformation)
                 self.__expireRequest(
                     statusInformation['sendPduHandle'],
                     self.__cache.pop(statusInformation['sendPduHandle']),
