@@ -46,20 +46,70 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         self.__timeline = {}
         self.__timelineExpQueue = {}
         self.__expirationTimer = 0
+        self.__paramsBranchId = -1
+
+    def __sec2usr(self, snmpEngine,  securityName, securityEngineID=None):
+        usmUserEngineID, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('SNMP-USER-BASED-SM-MIB', 'usmUserEngineID')
+        if self.__paramsBranchId != usmUserEngineID.branchVersionId:
+            usmUserName, usmUserSecurityName = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('SNMP-USER-BASED-SM-MIB', 'usmUserName', 'usmUserSecurityName')
+
+            self.__securityToUserMap = {}
+
+            nextMibNode = usmUserEngineID
+
+            while 1:
+                try:
+                    nextMibNode = usmUserEngineID.getNextNode(
+                                      nextMibNode.name
+                                  )
+                except NoSuchInstanceError:
+                    self.__paramsBranchId = usmUserEngineID.branchVersionId
+                    debug.logger & debug.flagSM and debug.logger('_sec2usr: built snmpEngineId + securityName to userName map, version %s: %r' % (self.__paramsBranchId, self.__securityToUserMap))
+                    break
+
+                instId = nextMibNode.name[len(usmUserSecurityName.name):]
+
+                __engineID = usmUserEngineID.getNode(usmUserEngineID.name + instId).syntax
+                __userName = usmUserName.getNode(usmUserName.name + instId).syntax
+                __securityName = usmUserSecurityName.getNode(usmUserSecurityName.name + instId).syntax
+
+                k = __engineID, __securityName
+
+                # first (lesser) securityName wins
+                if k not in self.__securityToUserMap:
+                    self.__securityToUserMap[k] = __userName
+
+        if securityEngineID is None:
+            snmpEngineID, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineID')
+            securityEngineID = snmpEngineID.syntax
+
+        try:
+            userName = self.__securityToUserMap[(securityEngineID, securityName)]
+        except KeyError:
+            debug.logger & debug.flagSM and debug.logger('_sec2usr: no entry exists for snmpEngineId %r, securityName %r' % (securityEngineID, securityName))
+            raise NoSuchInstanceError()  # emulate MIB lookup
+     
+        debug.logger & debug.flagSM and debug.logger('_sec2usr: using userName %r for snmpEngineId %r, securityName %r' % (userName, securityEngineID, securityName))
+
+        return userName
 
     def __getUserInfo(
-        self, mibInstrumController, securityEngineID, securityName
+        self, mibInstrumController, securityEngineID, userName
         ):
         usmUserEntry, = mibInstrumController.mibBuilder.importSymbols(
             'SNMP-USER-BASED-SM-MIB', 'usmUserEntry'
-            )
+        )
         tblIdx = usmUserEntry.getInstIdFromIndices(
-            securityEngineID, securityName
-            )
-        # Get protocols
+            securityEngineID, userName
+        )
+        # Get userName & securityName
+        usmUserName = usmUserEntry.getNode(
+            usmUserEntry.name + (2,) + tblIdx
+        ).syntax
         usmUserSecurityName = usmUserEntry.getNode(
             usmUserEntry.name + (3,) + tblIdx
-            ).syntax
+        ).syntax
+        # Get protocols
         usmUserAuthProtocol = usmUserEntry.getNode(
             usmUserEntry.name + (5,) + tblIdx
             ).syntax
@@ -77,73 +127,92 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             pysnmpUsmKeyEntry.name + (2,) + tblIdx
             ).syntax
         return (
-            usmUserSecurityName,  # XXX function needed?
+            usmUserName,
+            usmUserSecurityName,
             usmUserAuthProtocol,
             pysnmpUsmKeyAuthLocalized,
             usmUserPrivProtocol,
             pysnmpUsmKeyPrivLocalized
-            )
+        )
 
     def __cloneUserInfo(
-        self, mibInstrumController, securityEngineID, securityName
+            self, mibInstrumController, securityEngineID, userName
         ):
         snmpEngineID, = mibInstrumController.mibBuilder.importSymbols(
             '__SNMP-FRAMEWORK-MIB', 'snmpEngineID'
-            )
+        )
         # Proto entry
         usmUserEntry, = mibInstrumController.mibBuilder.importSymbols(
             'SNMP-USER-BASED-SM-MIB', 'usmUserEntry'
-            )
-        tblIdx = usmUserEntry.getInstIdFromIndices(
-            snmpEngineID.syntax, securityName
-            )
+        )
+        tblIdx1 = usmUserEntry.getInstIdFromIndices(
+            snmpEngineID.syntax, userName
+        )
         # Get proto protocols
+        usmUserName = usmUserEntry.getNode(
+            usmUserEntry.name + (2,) + tblIdx1
+        )
         usmUserSecurityName = usmUserEntry.getNode(
-            usmUserEntry.name + (3,) + tblIdx
-            )
+            usmUserEntry.name + (3,) + tblIdx1
+        )
+        usmUserCloneFrom = usmUserEntry.getNode(
+            usmUserEntry.name + (4,) + tblIdx1
+        )
         usmUserAuthProtocol = usmUserEntry.getNode(
-            usmUserEntry.name + (5,) + tblIdx
-            )
+            usmUserEntry.name + (5,) + tblIdx1
+        )
         usmUserPrivProtocol = usmUserEntry.getNode(
-            usmUserEntry.name + (8,) + tblIdx
-            )
+            usmUserEntry.name + (8,) + tblIdx1
+        )
         # Get proto keys
         pysnmpUsmKeyEntry, = mibInstrumController.mibBuilder.importSymbols(
             'PYSNMP-USM-MIB', 'pysnmpUsmKeyEntry'
-            )
+        )
         pysnmpUsmKeyAuth = pysnmpUsmKeyEntry.getNode(
-            pysnmpUsmKeyEntry.name + (3,) + tblIdx
-            )
+            pysnmpUsmKeyEntry.name + (3,) + tblIdx1
+        )
         pysnmpUsmKeyPriv = pysnmpUsmKeyEntry.getNode(
-            pysnmpUsmKeyEntry.name + (4,) + tblIdx
-            )        
+            pysnmpUsmKeyEntry.name + (4,) + tblIdx1
+        )        
         
         # Create new row from proto values
-        
-        tblIdx = usmUserEntry.getInstIdFromIndices(
-            securityEngineID, securityName
-            )
+
+        tblIdx2 = usmUserEntry.getInstIdFromIndices(
+            securityEngineID, userName
+        )
+
         # New row
         mibInstrumController.writeVars(
-            ((usmUserEntry.name + (13,) + tblIdx, 4),)
-            )
+            ((usmUserEntry.name + (13,) + tblIdx2, 4),)
+        )
+
+        # Set user&securityNames
+        usmUserEntry.getNode(
+            usmUserEntry.name + (2,) + tblIdx2
+        ).syntax = usmUserName.syntax
+        usmUserEntry.getNode(
+            usmUserEntry.name + (3,) + tblIdx2
+        ).syntax = usmUserSecurityName.syntax
+
+        # Store a reference to original row
+        usmUserEntry.getNode(
+            usmUserEntry.name + (4,) + tblIdx2
+        ).syntax = usmUserCloneFrom.syntax.clone(tblIdx1)
+ 
         # Set protocols
         usmUserEntry.getNode(
-            usmUserEntry.name + (3,) + tblIdx
-            ).syntax = usmUserSecurityName.syntax
+            usmUserEntry.name + (5,) + tblIdx2
+        ).syntax = usmUserAuthProtocol.syntax
         usmUserEntry.getNode(
-            usmUserEntry.name + (5,) + tblIdx
-            ).syntax = usmUserAuthProtocol.syntax
-        usmUserEntry.getNode(
-            usmUserEntry.name + (8,) + tblIdx
-            ).syntax = usmUserPrivProtocol.syntax
-        
+            usmUserEntry.name + (8,) + tblIdx2
+        ).syntax = usmUserPrivProtocol.syntax
+       
         # Localize and set keys
         pysnmpUsmKeyEntry, = mibInstrumController.mibBuilder.importSymbols(
             'PYSNMP-USM-MIB', 'pysnmpUsmKeyEntry'
             )
         pysnmpUsmKeyAuthLocalized = pysnmpUsmKeyEntry.getNode(
-            pysnmpUsmKeyEntry.name + (1,) + tblIdx
+            pysnmpUsmKeyEntry.name + (1,) + tblIdx2
             )
         if usmUserAuthProtocol.syntax in self.authServices:
             localizeKey = self.authServices[usmUserAuthProtocol.syntax].localizeKey
@@ -158,7 +227,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         if localAuthKey is not None:
             pysnmpUsmKeyAuthLocalized.syntax = pysnmpUsmKeyAuthLocalized.syntax.clone(localAuthKey)
         pysnmpUsmKeyPrivLocalized = pysnmpUsmKeyEntry.getNode(
-            pysnmpUsmKeyEntry.name + (2,) + tblIdx
+            pysnmpUsmKeyEntry.name + (2,) + tblIdx2
             )
         if usmUserPrivProtocol.syntax in self.privServices:
             localizeKey = self.privServices[usmUserPrivProtocol.syntax].localizeKey
@@ -174,12 +243,13 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         if localPrivKey is not None:
             pysnmpUsmKeyPrivLocalized.syntax = pysnmpUsmKeyPrivLocalized.syntax.clone(localPrivKey)
         return (
-            usmUserSecurityName.syntax,  # XXX function needed?
+            usmUserName.syntax,
+            usmUserSecurityName.syntax,
             usmUserAuthProtocol.syntax,
             pysnmpUsmKeyAuthLocalized.syntax,
             usmUserPrivProtocol.syntax,
             pysnmpUsmKeyPrivLocalized.syntax
-            )
+        )
               
     def __generateRequestOrResponseMsg(
         self,
@@ -200,10 +270,14 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             # 3.1.1a
             cachedSecurityData = self._cache.pop(securityStateReference)
             usmUserName = cachedSecurityData['msgUserName']
+            if 'usmUserSecurityName' in cachedSecurityData:
+                usmUserSecurityName = cachedSecurityData['usmUserSecurityName']
+            else:
+                usmUserSecurityName = usmUserName
             if 'usmUserAuthProtocol' in cachedSecurityData:
                 usmUserAuthProtocol = cachedSecurityData['usmUserAuthProtocol']
             else:
-                usmUserAuthProtocol = None
+                usmUserAuthProtocol = noauth.NoAuth.serviceID
             if 'usmUserAuthKeyLocalized' in cachedSecurityData:
                 usmUserAuthKeyLocalized = cachedSecurityData['usmUserAuthKeyLocalized']
             else:
@@ -211,7 +285,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             if 'usmUserPrivProtocol' in cachedSecurityData:
                 usmUserPrivProtocol = cachedSecurityData['usmUserPrivProtocol']
             else:
-                usmUserPrivProtocol = None
+                usmUserPrivProtocol = nopriv.NoPriv.serviceID
             if 'usmUserPrivKeyLocalized' in cachedSecurityData:
                 usmUserPrivKeyLocalized = cachedSecurityData['usmUserPrivKeyLocalized']
             else:
@@ -222,13 +296,15 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             # 3.1.1b
             try:
                 ( usmUserName,
+                  usmUserSecurityName,
                   usmUserAuthProtocol,
                   usmUserAuthKeyLocalized,
                   usmUserPrivProtocol,
                   usmUserPrivKeyLocalized ) = self.__getUserInfo(
                     snmpEngine.msgAndPduDsp.mibInstrumController,
-                    securityEngineID, securityName
-                    )
+                    securityEngineID,
+                    self.__sec2usr(snmpEngine, securityName, securityEngineID)
+                )
                 debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: read user info')
             except NoSuchInstanceError:
                 pysnmpUsmDiscovery, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__PYSNMP-USM-MIB', 'pysnmpUsmDiscovery')
@@ -236,14 +312,15 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                 if not __reportUnknownName:
                     try:
                         ( usmUserName,
+                          usmUserSecurityName,
                           usmUserAuthProtocol,
                           usmUserAuthKeyLocalized,
                           usmUserPrivProtocol,
                           usmUserPrivKeyLocalized ) = self.__cloneUserInfo(
                             snmpEngine.msgAndPduDsp.mibInstrumController,
                             securityEngineID,
-                            securityName
-                            )
+                            self.__sec2usr(snmpEngine, securityName)
+                        )
                     except NoSuchInstanceError:
                         __reportUnknownName = 1
 
@@ -252,27 +329,36 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                         errorIndication = errind.unknownSecurityName
                         )
                 debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: clone user info')
+            except PyAsn1Error:
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: %s' % (sys.exc_info()[1],))
+                snmpInGenErrs, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInGenErrs')
+                snmpInGenErrs.syntax = snmpInGenErrs.syntax + 1
+                raise error.StatusInformation(
+                    errorIndication=errind.invalidMsg
+                )
         else:
             # empty username used for engineID discovery
             usmUserName = usmUserSecurityName = null
-            usmUserAuthProtocol = usmUserAuthKeyLocalized = None
-            usmUserPrivProtocol = usmUserPrivKeyLocalized = None
+            usmUserAuthProtocol = noauth.NoAuth.serviceID
+            usmUserPrivProtocol = nopriv.NoPriv.serviceID
+            usmUserAuthKeyLocalized = usmUserPrivKeyLocalized = None
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: use empty USM data')
             
-        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: local user usmUserName %r usmUserAuthProtocol %s usmUserPrivProtocol %s securityEngineID %r securityName %r' % (usmUserName, usmUserAuthProtocol, usmUserPrivProtocol, securityEngineID, securityName))
+        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: local usmUserName %r usmUserSecurityName %r usmUserAuthProtocol %s usmUserPrivProtocol %s securityEngineID %r securityName %r' % (usmUserName, usmUserSecurityName, usmUserAuthProtocol, usmUserPrivProtocol, securityEngineID, securityName))
 
         msg = globalData
         
         # 3.1.2
         if securityLevel == 3:
-            if not usmUserAuthProtocol or not usmUserPrivProtocol:
+            if usmUserAuthProtocol == noauth.NoAuth.serviceID or \
+               usmUserPrivProtocol == nopriv.NoPriv.serviceID:
                 raise error.StatusInformation(
                     errorIndication = errind.unsupportedSecurityLevel
                     )
 
         # 3.1.3
         if securityLevel == 3 or securityLevel == 2:
-            if not usmUserAuthProtocol:
+            if usmUserAuthProtocol == noauth.NoAuth.serviceID:
                 raise error.StatusInformation(
                     errorIndication = errind.unsupportedSecurityLevel
                     )
@@ -321,29 +407,35 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             else:
                 raise error.StatusInformation(
                     errorIndication = errind.encryptionError
-                    )
+                )
 
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU %s' % scopedPDU.prettyPrint())
 
-            dataToEncrypt = encoder.encode(scopedPDU)
+            try:
+                dataToEncrypt = encoder.encode(scopedPDU)
+            except PyAsn1Error:
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU serialization error: %s' % sys.exc_info()[1])
+                raise error.StatusInformation(
+                    errorIndication = errind.serializationError
+                )
             
-            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU encoded into %r' % (dataToEncrypt,))
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU encoded into %s' % debug.hexdump(dataToEncrypt))
 
             ( encryptedData,
               privParameters ) = privHandler.encryptData(
                 usmUserPrivKeyLocalized,
                 ( snmpEngineBoots, snmpEngineTime, None ),
                 dataToEncrypt
-                )
+            )
 
             securityParameters.setComponentByPosition(
                 5, privParameters, verifyConstraints=False
-                )
+            )
             scopedPDUData.setComponentByPosition(
                 1, encryptedData, verifyConstraints=False
-                )
+            )
 
-            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU ciphered')
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU ciphered into %s' % debug.hexdump(encryptedData))
 
         # 3.1.4b
         elif securityLevel == 1 or securityLevel == 2:
@@ -354,18 +446,18 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         # 3.1.5
         securityParameters.setComponentByPosition(
             0, securityEngineID, verifyConstraints=False
-            )
+        )
         securityParameters.setComponentByPosition(
             1, snmpEngineBoots, verifyConstraints=False
-            )
+        )
         securityParameters.setComponentByPosition(
             2, snmpEngineTime, verifyConstraints=False
-            )
+        )
     
         # 3.1.7
         securityParameters.setComponentByPosition(
             3, usmUserName, verifyConstraints=False
-            )
+        )
 
         # 3.1.8a
         if securityLevel == 3 or securityLevel == 2:
@@ -383,17 +475,28 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
 
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: %s' % (securityParameters.prettyPrint(),))
             
-            msg.setComponentByPosition(
-                2, encoder.encode(securityParameters), verifyConstraints=False
+            try:
+                msg.setComponentByPosition(2, encoder.encode(securityParameters), verifyConstraints=False)
+            except PyAsn1Error:
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: securityParameters serialization error: %s' % sys.exc_info()[1])
+                raise error.StatusInformation(
+                    errorIndication = errind.serializationError
                 )
 
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: auth outgoing msg: %s' % msg.prettyPrint())
 
-            wholeMsg = encoder.encode(msg)
+            try:
+                wholeMsg = encoder.encode(msg)
+            except PyAsn1Error:
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: msg serialization error: %s' % sys.exc_info()[1])
+                raise error.StatusInformation(
+                    errorIndication = errind.serializationError
+                )
 
             authenticatedWholeMsg = authHandler.authenticateOutgoingMsg(
                 usmUserAuthKeyLocalized, wholeMsg
                 )
+
         # 3.1.8b
         else:
             securityParameters.setComponentByPosition(
@@ -402,13 +505,25 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
 
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: %s' % (securityParameters.prettyPrint(),))
 
-            msg.setComponentByPosition(
-                2, encoder.encode(securityParameters), verifyConstraints=False
+            try:
+                msg.setComponentByPosition(2, encoder.encode(securityParameters), verifyConstraints=False)
+            except PyAsn1Error:
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: secutiryParameters serialization error: %s' % sys.exc_info()[1])
+                raise error.StatusInformation(
+                    errorIndication = errind.serializationError
                 )
 
-            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: plain outgoing msg: %s' % msg.prettyPrint())
+            try:
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: plain outgoing msg: %s' % msg.prettyPrint())
+                authenticatedWholeMsg = encoder.encode(msg)
+            except PyAsn1Error:
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: msg serialization error: %s' % sys.exc_info()[1])
+                raise error.StatusInformation(
+                    errorIndication = errind.serializationError
+                )
+
+        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: %s outgoing msg: %s' % (securityLevel > 1 and "authenticated" or "plain", debug.hexdump(authenticatedWholeMsg)))
 
-            authenticatedWholeMsg = encoder.encode(msg)
         # 3.1.9
         return (
             msg.getComponentByPosition(2),
@@ -481,7 +596,9 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         # 3.2.9 -- moved up here to be able to report
         # maxSizeResponseScopedPDU on error
         # (48 - maximum SNMPv3 header length)
-        maxSizeResponseScopedPDU = maxMessageSize - len(securityParameters)-48
+        maxSizeResponseScopedPDU = int(maxMessageSize) - len(securityParameters) - 48
+
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: securityParameters %s' % debug.hexdump(securityParameters))
 
         # 3.2.1 
         try:
@@ -564,7 +681,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         if msgUserName:
             # 3.2.4
             try:
-                ( usmUserSecurityName,
+                ( usmUserName,
+                  usmUserSecurityName,
                   usmUserAuthProtocol,
                   usmUserAuthKeyLocalized,
                   usmUserPrivProtocol,
@@ -572,14 +690,15 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     snmpEngine.msgAndPduDsp.mibInstrumController,
                     msgAuthoritativeEngineID,
                     msgUserName
-                    )
+                )
                 debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read user info from LCD')
             except NoSuchInstanceError:
                 pysnmpUsmDiscoverable, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__PYSNMP-USM-MIB', 'pysnmpUsmDiscoverable')
                 __reportUnknownName = not pysnmpUsmDiscoverable.syntax
                 if not __reportUnknownName:
                     try:
-                        ( usmUserSecurityName,
+                        ( usmUserName,
+                          usmUserSecurityName,
                           usmUserAuthProtocol,
                           usmUserAuthKeyLocalized,
                           usmUserPrivProtocol,
@@ -587,63 +706,85 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                             snmpEngine.msgAndPduDsp.mibInstrumController,
                             msgAuthoritativeEngineID,
                             msgUserName
-                            )
+                        )
                         debug.logger & debug.flagSM and debug.logger('processIncomingMsg: cloned user info')
                     except NoSuchInstanceError:
                         __reportUnknownName = 1
                 debug.logger & debug.flagSM and debug.logger('processIncomingMsg: unknown securityEngineID %r msgUserName %r' % (msgAuthoritativeEngineID, msgUserName))
                 if __reportUnknownName:
-                        usmStatsUnknownUserNames, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnknownUserNames')
-                        usmStatsUnknownUserNames.syntax = usmStatsUnknownUserNames.syntax+1
-                        raise error.StatusInformation(
-                            errorIndication = errind.unknownSecurityName,
-                            oid = usmStatsUnknownUserNames.name,
-                            val = usmStatsUnknownUserNames.syntax,
-                            securityStateReference=securityStateReference,
-                            securityLevel=securityLevel,
-                            contextEngineId=contextEngineId,
-                            contextName=contextName,
-                            maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
-                            )
+                    usmStatsUnknownUserNames, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnknownUserNames')
+                    usmStatsUnknownUserNames.syntax = usmStatsUnknownUserNames.syntax+1
+                    raise error.StatusInformation(
+                        errorIndication = errind.unknownSecurityName,
+                        oid = usmStatsUnknownUserNames.name,
+                        val = usmStatsUnknownUserNames.syntax,
+                        securityStateReference=securityStateReference,
+                        securityLevel=securityLevel,
+                        contextEngineId=contextEngineId,
+                        contextName=contextName,
+                        maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
+                        )
+            except PyAsn1Error:
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: %s' % (sys.exc_info()[1],))
+                snmpInGenErrs, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInGenErrs')
+                snmpInGenErrs.syntax = snmpInGenErrs.syntax + 1
+                raise error.StatusInformation(
+                    errorIndication=errind.invalidMsg
+                )
         else:
             # empty username used for engineID discovery
             usmUserName = usmUserSecurityName = null
-            usmUserAuthProtocol = usmUserAuthKeyLocalized = None
-            usmUserPrivProtocol = usmUserPrivKeyLocalized = None
+            usmUserAuthProtocol = noauth.NoAuth.serviceID
+            usmUserPrivProtocol = nopriv.NoPriv.serviceID
+            usmUserAuthKeyLocalized = usmUserPrivKeyLocalized = None
 
-        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: now have usmUserSecurityName %s usmUserAuthProtocol %s usmUserPrivProtocol %s for msgUserName %s' % (usmUserSecurityName, usmUserAuthProtocol, usmUserPrivProtocol, msgUserName))
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: now have usmUserName %r usmUserSecurityName %r usmUserAuthProtocol %r usmUserPrivProtocol %r for msgUserName %r' % (usmUserName, usmUserSecurityName, usmUserAuthProtocol, usmUserPrivProtocol, msgUserName))
 
         # 3.2.11 (moved up here to let Reports be authenticated & encrypted)
         self._cache.pop(securityStateReference)
         securityStateReference = self._cache.push(
             msgUserName=securityParameters.getComponentByPosition(3),
+            usmUserSecurityName=usmUserSecurityName,
             usmUserAuthProtocol=usmUserAuthProtocol,
             usmUserAuthKeyLocalized=usmUserAuthKeyLocalized,
             usmUserPrivProtocol=usmUserPrivProtocol,
             usmUserPrivKeyLocalized=usmUserPrivKeyLocalized
-            )
+        )
 
         # 3.2.5
-        __reportError = 0
-        if securityLevel == 3:
-            if not usmUserAuthProtocol or not usmUserPrivProtocol:
-                __reportError = 1
+        if msgAuthoritativeEngineID == snmpEngineID:
+            # Authoritative SNMP engine: make sure securityLevel is sufficient
+            __badSecIndication = None
+            if securityLevel == 3:
+                if usmUserAuthProtocol == noauth.NoAuth.serviceID:
+                    __badSecIndication = 'authPriv wanted while auth not expected'
+                if usmUserPrivProtocol == nopriv.NoPriv.serviceID:
+                    __badSecIndication = 'authPriv wanted while priv not expected'
             elif securityLevel == 2:
-                if not usmUserAuthProtocol:
-                    __reportError = 1
-        if __reportError:
-            usmStatsUnsupportedSecLevels, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnsupportedSecLevels')
-            usmStatsUnsupportedSecLevels.syntax = usmStatsUnsupportedSecLevels.syntax + 1
-            raise error.StatusInformation(
-                errorIndication=errind.unsupportedSecurityLevel,
-                oid=usmStatsUnknownEngineIDs.name,
-                val=usmStatsUnknownEngineIDs.syntax,
-                securityStateReference=securityStateReference,
-                securityLevel=securityLevel,
-                contextEngineId=contextEngineId,
-                contextName=contextName,
-                maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
-                )
+                if usmUserAuthProtocol == noauth.NoAuth.serviceID:
+                    __badSecIndication = 'authNoPriv wanted while auth not expected'
+                if usmUserPrivProtocol != nopriv.NoPriv.serviceID:
+                    __badSecIndication = 'authNoPriv wanted while priv expected'
+
+            elif securityLevel == 1:
+                if usmUserAuthProtocol != noauth.NoAuth.serviceID:
+                    __badSecIndication = 'noAuthNoPriv wanted while auth expected'
+                if usmUserPrivProtocol != nopriv.NoPriv.serviceID:
+                    __badSecIndication = 'noAuthNoPriv wanted while priv expected'
+            if __badSecIndication:
+                usmStatsUnsupportedSecLevels, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnsupportedSecLevels')
+                usmStatsUnsupportedSecLevels.syntax = usmStatsUnsupportedSecLevels.syntax + 1
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: reporting inappropriate security level for user %s: %s' % (msgUserName, __badSecIndication))
+                raise error.StatusInformation(
+                    errorIndication=errind.unsupportedSecurityLevel,
+                    oid=usmStatsUnsupportedSecLevels.name,
+                    val=usmStatsUnsupportedSecLevels.syntax,
+                    securityStateReference=securityStateReference,
+                    securityLevel=securityLevel,
+                    contextEngineId=contextEngineId,
+                    contextName=contextName,
+                    maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
+                    )
 
         # 3.2.6
         if securityLevel == 3 or securityLevel == 2:
@@ -684,7 +825,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     int(time.time())
                     )
                 
-                expireAt = self.__expirationTimer + 300
+                expireAt = int(self.__expirationTimer + 300 / snmpEngine.transportDispatcher.getTimerResolution())
                 if expireAt not in self.__timelineExpQueue:
                     self.__timelineExpQueue[expireAt] = []
                 self.__timelineExpQueue[expireAt].append(
@@ -751,7 +892,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                         msgAuthoritativeEngineTime,
                         int(time.time())
                         )
-                    expireAt = self.__expirationTimer + 300
+                    expireAt = int(self.__expirationTimer + 300 / snmpEngine.transportDispatcher.getTimerResolution())
                     if expireAt not in self.__timelineExpQueue:
                         self.__timelineExpQueue[expireAt] = []
                     self.__timelineExpQueue[expireAt].append(
@@ -785,14 +926,14 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     )
 
             try:
-               decryptedData = privHandler.decryptData(
+                decryptedData = privHandler.decryptData(
                     usmUserPrivKeyLocalized,
-                    ( securityParameters.getComponentByPosition(1),
-                      securityParameters.getComponentByPosition(2),
-                      securityParameters.getComponentByPosition(5) ),
-                    encryptedPDU
+                        ( securityParameters.getComponentByPosition(1),
+                          securityParameters.getComponentByPosition(2),
+                          securityParameters.getComponentByPosition(5) ),
+                        encryptedPDU
                     )
-               debug.logger & debug.flagSM and debug.logger('processIncomingMsg: PDU deciphered into %r' % (decryptedData,))
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: PDU deciphered into %s' % debug.hexdump(decryptedData))
             except error.StatusInformation:
                 usmStatsDecryptionErrors, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsDecryptionErrors')
                 usmStatsDecryptionErrors.syntax = usmStatsDecryptionErrors.syntax+1
@@ -860,7 +1001,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             for engineIdKey in self.__timelineExpQueue[self.__expirationTimer]:
                 if engineIdKey in self.__timeline:
                     del self.__timeline[engineIdKey]
-                    debug.logger & debug.flagSM and debug.logger('__expireEnginesInfo: expiring %s' % (engineIdKey,))
+                    debug.logger & debug.flagSM and debug.logger('__expireTimelineInfo: expiring %r' % (engineIdKey,))
             del self.__timelineExpQueue[self.__expirationTimer]
         self.__expirationTimer = self.__expirationTimer + 1
         
