@@ -1,3 +1,5 @@
+import sys
+import traceback
 from pysnmp.smi.indices import OidOrderedDict
 from pysnmp.smi import exval, error
 from pysnmp.proto import rfc1902
@@ -226,27 +228,25 @@ OBJECT-TYPE\n\
      self.getReference())
         
 class MibTree(ObjectType):
-    branchVersionId = 0    # increments on tree structure change XXX
+    branchVersionId = 0    # cnanges on tree structure change
     maxAccess = 'not-accessible'
     def __init__(self, name, syntax=None):
         ObjectType.__init__(self, name, syntax)
-        self._vars = OidOrderedDict()            
+        self._vars = OidOrderedDict()
 
     # Subtrees registration
     
     def registerSubtrees(self, *subTrees):
-        """Register subtrees at this tree. Subtrees are always attached
-           at the level of this tree, not subtrees."""
+        self.branchVersionId += 1
         for subTree in subTrees:
             if subTree.name in self._vars:
                 raise error.SmiError(
                     'MIB subtree %s already registered at %s' %  (subTree.name, self)
                     )
             self._vars[subTree.name] = subTree
-            MibTree.branchVersionId = MibTree.branchVersionId + 1
 
     def unregisterSubtrees(self, *names):
-        """Detach subtrees from this tree"""
+        self.branchVersionId += 1
         for name in names:
             # This may fail if you fill a table by exporting MibScalarInstances
             # but later drop them through SNMP.
@@ -255,9 +255,14 @@ class MibTree(ObjectType):
                     'MIB subtree %s not registered at %s' %  (name, self)
                     )
             del self._vars[name]
-            MibTree.branchVersionId = MibTree.branchVersionId + 1
 
+    #
     # Tree traversal
+    #
+    # Missing branches are indicated by the NoSuchObjectError exception.
+    # Although subtrees may indicate their missing branches by the 
+    # NoSuchInstanceError exception.
+    #
 
     def getBranch(self, name, idx):
         """Return a branch of this tree where the 'name' OID may reside"""
@@ -268,6 +273,17 @@ class MibTree(ObjectType):
                 
         raise error.NoSuchObjectError(name=name, idx=idx)
 
+    def getNextBranch(self, name, idx=None):
+        # Start from the beginning
+        if self._vars: first = list(self._vars.keys())[0]
+        if self._vars and name < first:
+            return self._vars[first]
+        else:
+            try:
+                return self._vars[self._vars.nextKey(name)]
+            except KeyError:
+                raise error.NoSuchObjectError(idx=idx, name=name)
+ 
     def getNode(self, name, idx=None):
         """Return tree node found by name"""
         if name == self.name:
@@ -279,26 +295,17 @@ class MibTree(ObjectType):
         """Return tree node next to name"""
         try:
             nextNode = self.getBranch(name, idx)
-        except error.NoSuchObjectError:
-            # Start from the beginning
-            if self._vars: first = list(self._vars.keys())[0]
-            if self._vars and name <= first:
-                return self._vars[first]
-            else:
-                # Try following the white rabbit at our level
-                try:
-                    return self._vars[self._vars.nextKey(name)]
-                except KeyError:
-                    raise error.NoSuchObjectError(idx=idx, name=name)
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            return self.getNextBranch(name, idx)
         else:
             try:
                 return nextNode.getNextNode(name, idx)
-            except error.NoSuchObjectError:
+            except (error.NoSuchInstanceError, error.NoSuchObjectError):
                 try:
                     return self._vars[self._vars.nextKey(nextNode.name)]
                 except KeyError:
                     raise error.NoSuchObjectError(idx=idx, name=name)
-                
+ 
     # MIB instrumentation
 
     # Read operation
@@ -314,49 +321,80 @@ class MibTree(ObjectType):
         else:
             try:
                 node = self.getBranch(name, idx)
-            except error.NoSuchObjectError:
+            except (error.NoSuchInstanceError, error.NoSuchObjectError):
                 return # missing object is not an error here
-
-            node.readTest(name, val, idx, (acFun, acCtx))
+            else:
+                node.readTest(name, val, idx, acInfo)
         
     def readGet(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         try:
             node = self.getBranch(name, idx)
-        except error.NoSuchObjectError:
-            return name, exval.noSuchInstance
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            return name, exval.noSuchObject
         else:
-            return node.readGet(name, val, idx, (acFun, acCtx))
+            return node.readGet(name, val, idx, acInfo)
 
     # Read next operation is subtree-specific
-    
-    def readTestNext(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
+   
+    depthFirst, breadthFirst = 0, 1
+     
+    def readTestNext(self, name, val, idx, acInfo, oName=None):
+        if oName is None:
+            oName = name
+            topOfTheMib = True
+        else:
+            topOfTheMib = False
         nextName = name
+        direction = self.depthFirst
         while 1:  # XXX linear search here
-            try:
-                nextName = self.getNextNode(nextName, idx).name
-            except error.NoSuchObjectError:
-                return # missing object is not an error here
-            try:
-                return self.readTest(nextName, val, idx, (acFun, acCtx))
-            except error.NoAccessError:
-                continue
-    
-    def readGetNext(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
-        nextName = name
-        while 1:
-            try:
-                nextName = self.getNextNode(nextName, idx).name
-            except error.NoSuchObjectError:
-                return name, exval.endOfMib
-            try:
-                self.readTest(nextName, val, idx, (acFun, acCtx)) # XXX
-            except error.NoAccessError:
-                continue
+            if direction == self.depthFirst:
+                direction = self.breadthFirst
+                try:
+                    node = self.getBranch(nextName, idx)
+                except (error.NoSuchInstanceError, error.NoSuchObjectError):
+                    continue
             else:
-                return self.readGet(nextName, val, idx, (acFun, acCtx))
+                try:                
+                    node = self.getNextBranch(nextName, idx)
+                except (error.NoSuchInstanceError, error.NoSuchObjectError):
+                    if topOfTheMib:
+                        return
+                    raise
+                direction = self.depthFirst
+                nextName = node.name
+            try:
+                return node.readTestNext(nextName, val, idx, acInfo, oName)
+            except (error.NoAccessError, error.NoSuchInstanceError, error.NoSuchObjectError):
+                pass
+    
+    def readGetNext(self, name, val, idx, acInfo, oName=None):
+        if oName is None:
+            oName = name
+            topOfTheMib = True
+        else:
+            topOfTheMib = False
+        nextName = name
+        direction = self.depthFirst
+        while 1:  # XXX linear search here
+            if direction == self.depthFirst:
+                direction = self.breadthFirst
+                try:
+                    node = self.getBranch(nextName, idx)
+                except (error.NoSuchInstanceError, error.NoSuchObjectError):
+                    continue
+            else:
+                try:
+                    node = self.getNextBranch(nextName, idx)
+                except (error.NoSuchInstanceError, error.NoSuchObjectError):
+                    if topOfTheMib:
+                        return name, exval.endOfMib
+                    raise
+                direction = self.depthFirst
+                nextName = node.name
+            try:
+                return node.readGetNext(nextName, val, idx, acInfo, oName)
+            except (error.NoAccessError, error.NoSuchInstanceError, error.NoSuchObjectError):
+                pass
 
     # Write operation
     
@@ -370,54 +408,107 @@ class MibTree(ObjectType):
                     raise error.NotWritableError(idx=idx, name=name)
         else:
             node = self.getBranch(name, idx)
-            node.writeTest(name, val, idx, (acFun, acCtx))
+            node.writeTest(name, val, idx, acInfo)
     
     def writeCommit(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
-        self.getBranch(name, idx).writeCommit(name, val, idx, (acFun, acCtx))
+        self.getBranch(name, idx).writeCommit(name, val, idx, acInfo)
     
     def writeCleanup(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
-        self.getBranch(name, idx).writeCleanup(name, val, idx, (acFun, acCtx))
+        self.branchVersionId += 1
+        self.getBranch(name, idx).writeCleanup(name, val, idx, acInfo)
     
     def writeUndo(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
-        self.getBranch(name, idx).writeUndo(name, val, idx, (acFun, acCtx))
+        self.getBranch(name, idx).writeUndo(name, val, idx, acInfo)
 
 class MibScalar(MibTree):
     """Scalar MIB variable. Implements access control checking."""
     maxAccess = 'readonly'
-        
+
+    #
+    # Subtree traversal
+    #
+    # Missing branches are indicated by the NoSuchInstanceError exception.
+    #
+
+    def getBranch(self, name, idx):
+        try:
+            return MibTree.getBranch(self, name, idx)
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+
+    def getNextBranch(self, name, idx=None):
+        try:
+            return MibTree.getNextBranch(self, name, idx)
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+
+    def getNode(self, name, idx=None):
+        try:
+            return MibTree.getNode(self, name, idx)
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+
+    def getNextNode(self, name, idx=None):
+        try:
+            return MibTree.getNextNode(self, name, idx)
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            raise error.NoSuchInstanceError(idx=idx, name=name)       
+
     # MIB instrumentation methods
     
     # Read operation
-    
+ 
     def readTest(self, name, val, idx, acInfo):
         (acFun, acCtx) = acInfo
         if name == self.name:
             raise error.NoAccessError(idx=idx, name=name)
-        else:
-            MibTree.readTest(self, name, val, idx, (acFun, acCtx))
-        # If instance exists, check permissions
         if acFun:
             if self.maxAccess not in (
                 'readonly', 'readwrite', 'readcreate'
                 ) or acFun(name, self.syntax, idx, 'read', acCtx):
                 raise error.NoAccessError(idx=idx, name=name)
-    
+        MibTree.readTest(self, name, val, idx, acInfo)
+
+    def readGet(self, name, val, idx, acInfo):
+        try:
+            node = self.getBranch(name, idx)
+        except error.NoSuchInstanceError:
+            return name, exval.noSuchInstance
+        else:
+            return node.readGet(name, val, idx, acInfo)
+
+    def readTestNext(self, name, val, idx, acInfo, oName=None):
+        (acFun, acCtx) = acInfo
+        if acFun:
+            if self.maxAccess not in (
+                'readonly', 'readwrite', 'readcreate'
+                ) or acFun(name, self.syntax, idx, 'read', acCtx):
+                raise error.NoAccessError(idx=idx, name=name)
+        MibTree.readTestNext(self, name, val, idx, acInfo, oName)
+ 
+    def readGetNext(self, name, val, idx, acInfo, oName=None):
+        (acFun, acCtx) = acInfo
+        # have to dublicate AC here as *Next code above treats
+        # noAccess as a noSuchObject at the Test stage, goes on
+        # to Reading
+        if acFun:
+            if self.maxAccess not in (
+                'readonly', 'readwrite', 'readcreate'
+                ) or acFun(name, self.syntax, idx, 'read', acCtx):
+                raise error.NoAccessError(idx=idx, name=name)
+        return MibTree.readGetNext(self, name, val, idx, acInfo, oName)
+ 
     # Two-phase commit implementation
 
     def writeTest(self, name, val, idx, acInfo):
         (acFun, acCtx) = acInfo
         if name == self.name:
             raise error.NoAccessError(idx=idx, name=name)
-        else:
-            MibTree.writeTest(self, name, val, idx, (acFun, acCtx))
-        # If instance exists, check permissions
         if acFun:
             if self.maxAccess not in ('readwrite', 'readcreate') or \
                    acFun(name, self.syntax, idx, 'write', acCtx):
                 raise error.NotWritableError(idx=idx, name=name)
+        MibTree.writeTest(self, name, val, idx, acInfo)
 
 class MibScalarInstance(MibTree):
     """Scalar MIB variable instance. Implements read/write operations."""
@@ -426,7 +517,47 @@ class MibScalarInstance(MibTree):
         self.typeName = typeName
         self.instId = instId
         self.__oldSyntax = None
-        
+
+    #
+    # Managed object value access methods
+    #
+
+    def getValue(self, name, idx):
+        debug.logger & debug.flagIns and debug.logger('getValue: returning %r for %s' % (self.syntax, self.name))
+        return self.syntax.clone()
+
+    def setValue(self, value, name, idx):
+        try:
+            if hasattr(self.syntax, 'setValue'):
+                return self.syntax.setValue(value)
+            else:
+                return self.syntax.clone(value)
+        except PyAsn1Error:
+            exc_t, exc_v, exc_tb = sys.exc_info()
+            debug.logger & debug.flagIns and debug.logger('setValue: %s=%r failed with traceback %s' % (self.name, value, traceback.format_exception(exc_t, exc_v, exc_tb)))
+            if isinstance(exc_v, error.TableRowManagement):
+                raise exc_v
+            else:
+                raise error.WrongValueError(idx=idx, name=name, msg=exc_v)
+
+    #
+    # Subtree traversal
+    #
+    # Missing branches are indicated by the NoSuchInstanceError exception.
+    #
+
+    def getBranch(self, name, idx):
+        try:
+            return MibTree.getBranch(self, name, idx)
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+
+    def getNextBranch(self, name, idx=None):
+        try:
+            return MibTree.getNextBranch(self, name, idx)
+        except (error.NoSuchInstanceError, error.NoSuchObjectError):
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+
     def getNode(self, name, idx=None):
         # Recursion terminator
         if name == self.name:
@@ -442,30 +573,44 @@ class MibScalarInstance(MibTree):
     
     def readTest(self, name, val, idx, acInfo):
         if name != self.name:
-            raise error.NoSuchObjectError(idx=idx, name=name)
+            raise error.NoSuchInstanceError(idx=idx, name=name)
 
     def readGet(self, name, val, idx, acInfo):
-        # Return current variable (name, value). This is the only API method
-        # capable of returning anything!
+        # Return current variable (name, value). 
         if name == self.name:
             debug.logger & debug.flagIns and debug.logger('readGet: %s=%r' % (self.name, self.syntax))
-            return self.name, self.syntax.clone()
+            return self.name, self.getValue(name, idx)
         else:
-            raise error.NoSuchObjectError(idx=idx, name=name)
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+ 
+    def readTestNext(self, name, val, idx, acInfo, oName=None):
+        if name != self.name or name <= oName:
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+
+    def readGetNext(self, name, val, idx, acInfo, oName=None):
+        if name == self.name and name > oName:
+            debug.logger & debug.flagIns and debug.logger('readGetNext: %s=%r' % (self.name, self.syntax))
+            return self.readGet(name, val, idx, acInfo)
+        else:
+            raise error.NoSuchInstanceError(idx=idx, name=name)
     
     # Write operation: two-phase commit
 
     def writeTest(self, name, val, idx, acInfo):
         # Make sure write's allowed
         if name == self.name:
-            if hasattr(self.syntax, 'smiWrite'):
-                self.__newSyntax = self.syntax.smiWrite(name, val, idx)
-            else:
-                self.__newSyntax = self.syntax.clone(val)
-            if hasattr(self.__newSyntax, 'smiRaisePendingError'):
-                self.__newSyntax.smiRaisePendingError()
+            try:
+                self.__newSyntax = self.setValue(val, name, idx)
+            except error.MibOperationError:
+                # SMI exceptions may carry additional content
+                why = sys.exc_info()[1]
+                if 'syntax' in why:
+                    self.__newSyntax = why['syntax']
+                    raise why
+                else:
+                    raise error.WrongValueError(idx=idx, name=name, msg=sys.exc_info()[1])
         else:
-            raise error.NoSuchObjectError(idx=idx, name=name)
+            raise error.NoSuchInstanceError(idx=idx, name=name)
 
     def writeCommit(self, name, val, idx, acInfo):
         # Backup original value
@@ -475,6 +620,7 @@ class MibScalarInstance(MibTree):
         self.syntax = self.__newSyntax
         
     def writeCleanup(self, name, val, idx, acInfo):
+        self.branchVersionId += 1
         debug.logger & debug.flagIns and debug.logger('writeCleanup: %s=%r' % (name, val))
         # Drop previous value
         self.__newSyntax = self.__oldSyntax = None
@@ -490,38 +636,48 @@ class MibScalarInstance(MibTree):
 
     def createTest(self, name, val, idx, acInfo):
         if name == self.name:
-            if hasattr(self.syntax, 'smiCreate'):
-                self.__newSyntax = self.syntax.smiCreate(name, val, idx)
-            else:
-                self.__newSyntax = self.syntax.clone(val)
+            try:
+                self.__newSyntax = self.setValue(val, name, idx)
+            except error.MibOperationError:
+                # SMI exceptions may carry additional content
+                why = sys.exc_info()[1]
+                if 'syntax' in why:
+                    self.__newSyntax = why['syntax']
+                else:
+                    raise error.WrongValueError(idx=idx, name=name, msg=sys.exc_info()[1])
         else:
-            raise error.NoSuchObjectError(idx=idx, name=name)
+            raise error.NoSuchInstanceError(idx=idx, name=name)
+
     def createCommit(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         if val is not None:
-            self.writeCommit(name, val, idx, (acFun, acCtx))
+            self.writeCommit(name, val, idx, acInfo)
+
     def createCleanup(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
+        self.branchVersionId += 1
         debug.logger & debug.flagIns and debug.logger('createCleanup: %s=%r' % (name, val))
         if val is not None:
-            self.writeCleanup(name, val, idx, (acFun, acCtx))
+            self.writeCleanup(name, val, idx, acInfo)
+
     def createUndo(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         if val is not None:
-            self.writeUndo(name, val, idx, (acFun, acCtx))
+            self.writeUndo(name, val, idx, acInfo)
 
     # Destroy operation
 
     def destroyTest(self, name, val, idx, acInfo):
         if name == self.name:
-            if hasattr(self.syntax, 'smiDestroy'):
-                self.__newSyntax = self.syntax.smiDestoy(name, val)
-            else:
-                self.__newSyntax = self.syntax.clone(val)
+            try:
+                self.__newSyntax = self.setValue(val, name, idx)
+            except error.MibOperationError:
+                # SMI exceptions may carry additional content
+                why = sys.exc_info()[1]
+                if 'syntax' in why:
+                    self.__newSyntax = why['syntax']
         else:
-            raise error.NoSuchObjectError(idx=idx, name=name)
+            raise error.NoSuchInstanceError(idx=idx, name=name)
     def destroyCommit(self, name, val, idx, acInfo): pass
-    def destroyCleanup(self, name, val, idx, acInfo): pass
+    def destroyCleanup(self, name, val, idx, acInfo):
+        self.branchVersionId += 1
     def destroyUndo(self, name, val, idx, acInfo): pass
 
 # Conceptual table classes
@@ -533,60 +689,21 @@ class MibTableColumn(MibScalar):
         MibScalar.__init__(self, name, syntax)
         self.__createdInstances = {}; self.__destroyedInstances = {}
         self.__rowOpWanted = {}
-        self.__valIdx = {} # column instance value to OID index
-        self.__valIdxId = -1
-        
-    # No branches here, terminal OIDs only
+    
+    #
+    # Subtree traversal
+    #
+    # Missing leaves are indicated by the NoSuchInstanceError exception.
+    #
+
     def getBranch(self, name, idx):
         if name in self._vars:
             return self._vars[name]
-        raise error.NoSuchObjectError(name=name, idx=idx)
-
-    def getNode(self, name, idx=None):
-        try:
-            return MibScalar.getNode(self, name, idx=None)
-        except error.NoSuchObjectError:
-            raise error.NoSuchInstanceError(idx=idx, name=name)
-
-    def getNextNode(self, name, idx=None):
-        try:
-            return MibScalar.getNextNode(self, name, idx=None)
-        except error.NoSuchObjectError:
-            raise error.NoSuchInstanceError(idx=idx, name=name)
+        raise error.NoSuchInstanceError(name=name, idx=idx)
 
     def setProtoInstance(self, protoInstance):
         self.protoInstance = protoInstance
 
-    # Value to instances map
-
-    def getNextNodeWithValue(self, name, value):
-        if not self.__valIdx or self.__valIdxId != self.branchVersionId:
-            # Build a value->oid index
-            node = self
-            while 1:
-                try:
-                    node = self.getNextNode(node.name)
-                except error.NoSuchInstanceError:
-                    break
-                if node.syntax not in self.__valIdx:
-                    self.__valIdx[node.syntax] = OidOrderedDict()
-                self.__valIdx[node.syntax][node.name] = 1
-
-            debug.logger & debug.flagIns and debug.logger('value index rebuilt at %s, %s entries' % (self.name,len(self.__valIdx)))
-
-            # Sync to tree version
-            self.__valIdxId = self.branchVersionId
-
-        if value in self.__valIdx:
-            try:
-                return self.getNode(
-                    self.__valIdx[value].nextKey(name)
-                    )
-            except KeyError:
-                raise error.NoSuchInstanceError(name=name)
-        else:
-            raise error.NoSuchInstanceError(name=name)
-            
     # Column creation (this should probably be converted into some state
     # machine for clarity). Also, it might be a good idea to inidicate
     # defaulted cols creation in a clearer way than just a val == None.
@@ -609,41 +726,32 @@ class MibTableColumn(MibScalar):
         self.__createdInstances[name] = self.protoInstance(
             self.name, name[len(self.name):], self.syntax.clone()
             )
-        self.__createdInstances[name].createTest(
-            name, val, idx, (acFun, acCtx)
-            )
+        self.__createdInstances[name].createTest(name, val, idx, acInfo)
             
     def createCommit(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Commit new instance value
         if name in self._vars: # XXX
             if name in self.__createdInstances:
-                self._vars[name].createCommit(name, val, idx, (acFun, acCtx))
+                self._vars[name].createCommit(name, val, idx, acInfo)
             return
-        self.__createdInstances[name].createCommit(
-            name, val, idx, (acFun, acCtx)
-            )
+        self.__createdInstances[name].createCommit(name, val, idx, acInfo)
         # ...commit new column instance
         self._vars[name], self.__createdInstances[name] = \
                           self.__createdInstances[name], self._vars.get(name)
 
     def createCleanup(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
-        # Drop by-value index
-        self.__valIdx.clear()
-        
         # Drop previous column instance
+        self.branchVersionId += 1
         if name in self.__createdInstances:
             if self.__createdInstances[name] is not None:
                 self.__createdInstances[name].createCleanup(
-                    name, val, idx, (acFun, acCtx)
+                    name, val, idx, acInfo
                     )
             del self.__createdInstances[name]
         elif name in self._vars:
-            self._vars[name].createCleanup(name, val, idx, (acFun, acCtx))
+            self._vars[name].createCleanup(name, val, idx, acInfo)
 
     def createUndo(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Set back previous column instance, drop the new one
         if name in self.__createdInstances:
             self._vars[name] = self.__createdInstances[name]
@@ -658,7 +766,7 @@ class MibTableColumn(MibScalar):
                 except PyAsn1Error:
                     del self._vars[name]
                 else:
-                    self._vars[name].createUndo(name, val, idx, (acFun, acCtx))
+                    self._vars[name].createUndo(name, val, idx, acInfo)
                 
     # Column destruction
         
@@ -673,95 +781,80 @@ class MibTableColumn(MibScalar):
             if val is not None and self.maxAccess != 'readcreate' or \
                    acFun(name, self.syntax, idx, 'write', acCtx):
                 raise error.NoAccessError(idx=idx, name=name)
-        self._vars[name].destroyTest(
-            name, val, idx, (acFun, acCtx)
-            )
+        self._vars[name].destroyTest(name, val, idx, acInfo)
 
     def destroyCommit(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Make a copy of column instance and take it off the tree
         if name in self._vars:
-            self._vars[name].destroyCommit(
-                name, val, idx, (acFun, acCtx)
-                )            
+            self._vars[name].destroyCommit(name, val, idx, acInfo)          
             self.__destroyedInstances[name] = self._vars[name]
             del self._vars[name]
         
     def destroyCleanup(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
-        # Drop by-value index
-        self.__valIdx.clear()
-        
         # Drop instance copy
+        self.branchVersionId += 1
         if name in self.__destroyedInstances:
             self.__destroyedInstances[name].destroyCleanup(
-                name, val, idx, (acFun, acCtx)
+                name, val, idx, acInfo
                 )
             debug.logger & debug.flagIns and debug.logger('destroyCleanup: %s=%r' % (name, val))
             del self.__destroyedInstances[name]
             
     def destroyUndo(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Set back column instance
         if name in self.__destroyedInstances:
             self._vars[name] = self.__destroyedInstances[name]
             self._vars[name].destroyUndo(
-                name, val, idx, (acFun, acCtx)
+                name, val, idx, acInfo
                 )            
             del self.__destroyedInstances[name]
             
     # Set/modify column
 
     def writeTest(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Besides common checks, request row creation on no-instance
         try:
             # First try the instance
             MibScalar.writeTest(
-                self, name, val, idx, (acFun, acCtx)
+                self, name, val, idx, acInfo
                 )
         # ...otherwise proceed with creating new column
-        except (error.NoSuchObjectError, error.RowCreationWanted):
+        except (error.NoSuchInstanceError, error.RowCreationWanted):
             self.__rowOpWanted[name] = error.RowCreationWanted()
-            self.createTest(name, val, idx, (acFun, acCtx))
+            self.createTest(name, val, idx, acInfo)
         except error.RowDestructionWanted:
             self.__rowOpWanted[name] = error.RowDestructionWanted()
-            self.destroyTest(name, val, idx, (acFun, acCtx))
+            self.destroyTest(name, val, idx, acInfo)
         if name in self.__rowOpWanted:
-            debug.logger & debug.flagIns and debug.logger('%s flagged by %s=%r' % (self.__rowOpWanted[name], name, val))
+            debug.logger & debug.flagIns and debug.logger('%s flagged by %s=%r, exception %s' % (self.__rowOpWanted[name], name, val, sys.exc_info()[1]))
             raise self.__rowOpWanted[name]
 
     def __delegateWrite(self, subAction, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         if name not in self.__rowOpWanted:
             getattr(MibScalar, 'write'+subAction)(
-                self, name, val, idx, (acFun, acCtx)
+                self, name, val, idx, acInfo
                 )
             return
         if isinstance(self.__rowOpWanted[name], error.RowCreationWanted):
             getattr(self, 'create'+subAction)(
-                name, val, idx, (acFun, acCtx)
+                name, val, idx, acInfo
                 )
         if isinstance(self.__rowOpWanted[name], error.RowDestructionWanted):
             getattr(self, 'destroy'+subAction)(
-                name, val, idx, (acFun, acCtx)
+                           name, val, idx, acInfo
                 )
         
     def writeCommit(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         self.__delegateWrite(
-            'Commit', name, val, idx, (acFun, acCtx)
+            'Commit', name, val, idx, acInfo
             )
         if name in self.__rowOpWanted:
             raise self.__rowOpWanted[name]
 
     def writeCleanup(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
-        # Drop by-value index
-        self.__valIdx.clear()
-        
+        self.branchVersionId += 1
         self.__delegateWrite(
-            'Cleanup', name, val, idx, (acFun, acCtx)
+            'Cleanup', name, val, idx, acInfo
             )
         if name in self.__rowOpWanted:
             e = self.__rowOpWanted[name]
@@ -770,9 +863,8 @@ class MibTableColumn(MibScalar):
             raise e
             
     def writeUndo(self, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         self.__delegateWrite(
-            'Undo', name, val, idx, (acFun, acCtx)
+            'Undo', name, val, idx, acInfo
             )
         if name in self.__rowOpWanted:
             e = self.__rowOpWanted[name]
@@ -857,7 +949,6 @@ class MibTableRow(MibTree):
     # Fate sharing mechanics
 
     def announceManagementEvent(self, action, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Convert OID suffix into index vals
         instId = name[len(self.name)+1:]
         baseIndices = []
@@ -879,11 +970,10 @@ class MibTableRow(MibTree):
             mibObj, = mibBuilder.importSymbols(modName, mibSym)
             debug.logger & debug.flagIns and debug.logger('announceManagementEvent %s to %s' % (action, mibObj))
             mibObj.receiveManagementEvent(
-                action, baseIndices, val, idx, (acFun, acCtx)
+                action, baseIndices, val, idx, acInfo
                 )
             
     def receiveManagementEvent(self, action, baseIndices, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # The default implementation supports one-to-one rows dependency
         newSuffix = ()
         # Resolve indices intersection
@@ -894,8 +984,7 @@ class MibTableRow(MibTree):
                     newSuffix = newSuffix + self.getAsName(syntax, impliedFlag)
         if newSuffix:
             debug.logger & debug.flagIns and debug.logger('receiveManagementEvent %s for suffix %s' % (action, newSuffix))
-            self.__manageColumns(action, (), newSuffix, val, idx,
-                                 (acFun, acCtx))
+            self.__manageColumns(action, (), newSuffix, val, idx, acInfo)
 
     def registerAugmentions(self, *names):
         for modName, symName in names:
@@ -917,7 +1006,6 @@ class MibTableRow(MibTree):
                              
     def __manageColumns(self, action, excludeName, nameSuffix,
                         val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Build a map of index names and values for automatic initialization
         indexVals = {}; instId = nameSuffix
         for impliedFlag, modName, symName in self.indexNames:
@@ -933,32 +1021,30 @@ class MibTableRow(MibTree):
                 getattr(var, action)(name + nameSuffix, indexVals[name], idx,
                                      (None, None))
             else:
-                getattr(var, action)(name + nameSuffix, val, idx,
-                                     (acFun, acCtx))
+                getattr(var, action)(name + nameSuffix, val, idx, acInfo)
             debug.logger & debug.flagIns and debug.logger('__manageColumns: action %s name %s suffix %s %svalue %r' % (action, name, nameSuffix, name in indexVals and "index " or "", indexVals.get(name, val)))
 
     def __delegate(self, subAction, name, val, idx, acInfo):
-        (acFun, acCtx) = acInfo
         # Relay operation request to column, expect row operation request.
         try:
             getattr(self.getBranch(name, idx), 'write'+subAction)(
-                name, val, idx, (acFun, acCtx)
+                name, val, idx, acInfo
                 )
         except error.RowCreationWanted:
             self.__manageColumns(
                 'create'+subAction, name[:len(self.name)+1],
-                name[len(self.name)+1:], None, idx, (acFun, acCtx)
+                name[len(self.name)+1:], None, idx, acInfo
                 )
             self.announceManagementEvent(
-                'create'+subAction, name, None, idx, (acFun, acCtx)
+                'create'+subAction, name, None, idx, acInfo
                 )
         except error.RowDestructionWanted:
             self.__manageColumns(
                 'destroy'+subAction, name[:len(self.name)+1],
-                name[len(self.name)+1:], None, idx, (acFun, acCtx)
+                name[len(self.name)+1:], None, idx, acInfo
                 )
             self.announceManagementEvent(
-                'destroy'+subAction, name, None, idx, (acFun,acCtx)
+                'destroy'+subAction, name, None, idx, acInfo
                 )
     
     def writeTest(self, name, val, idx, acInfo):
@@ -966,6 +1052,7 @@ class MibTableRow(MibTree):
     def writeCommit(self, name, val, idx, acInfo):
         self.__delegate('Commit', name, val, idx, acInfo)
     def writeCleanup(self, name, val, idx, acInfo):
+        self.branchVersionId += 1
         self.__delegate('Cleanup', name, val, idx, acInfo)
     def writeUndo(self, name, val,  idx, acInfo):
         self.__delegate('Undo', name, val, idx, acInfo)
@@ -1000,7 +1087,7 @@ class MibTableRow(MibTree):
     def getInstIdFromIndices(self, *indices):
         """Return column instance identification from indices"""
         if indices in self.__idxToIdCache:
-          return self.__idxToIdCache[indices]
+            return self.__idxToIdCache[indices]
         idx = 0; idxLen = len(indices); instId = ()
         for impliedFlag, modName, symName in self.indexNames:
             mibObj, = mibBuilder.importSymbols(modName, symName)
